###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import re
from collections import defaultdict

from biolib.taxonomy import Taxonomy

from gtdb_validation_tk.common import (parse_species_ledgers,
                                        parse_genus_ledger,
                                        parse_gtdb_synonym_file,
                                        gtdb_canonical_species_id,
                                        canonical_taxon_name,
                                        canonical_species_name,
                                        canonical_wgs_id,
                                        latinized_generic_sp_name,
                                        latinized_specific_sp_name)


class CheckSpeciesNames(object):
    """Validate species names."""

    def __init__(self):
        """Initialize."""
        
        self.logger = logging.getLogger()
        
    def read_user_genome_table(self, user_genome_id_table):
        """Read user genome ID table."""
        
        uba_map = {}
        ncbi_map = {}
        for line in open(user_genome_id_table):
            line_split = line.strip().split('\t')
            if len(line_split) == 3:
                ncbi_map[line_split[0]] = line_split[2]
                uba_map[line_split[0]] = line_split[1]
            else:
                uba_map[line_split[0]] = line_split[1]
                
        return uba_map, ncbi_map
        
    def gtdb_species_id(self, gid, specific_name):
        """Check if specific name has the form sp<accession>."""
        
        canonical_name = canonical_taxon_name(sspecific_name)
        
        if canonical_name.startswith('sp'):
            sp_id = gid[gid.find('_')+1:gid.rfind('.')]
            if canonical_name[2:] == sp_id:
                return True
                
        return False

    def check_generic(self,
                        input_taxonomy,
                        gtdb_metadata_file,
                        sp_classification_ledger,
                        genus_ledger,
                        gtdb_synonym_file,
                        user_genome_id_table,
                        output_dir):
        """Validate generic name of species."""
        
        # read user genome ID table
        uba_map, ncbi_map = self.read_user_genome_table(user_genome_id_table)
        
        # read species ledger
        species_ledger = parse_species_ledgers(sp_classification_ledger, None)
        
        # read genus ledger
        genus_ledger = parse_genus_ledger(genus_ledger)
        
        # get genomes in each GTDB genus
        gtdb_genera = defaultdict(list)
        gtdb_taxonomy = {}
        for line in open(input_taxonomy):
            line_split = line.strip().split('\t')
            
            gid = line_split[0]
            taxa = [t.strip() for t in line_split[1].split(';')]
            gtdb_genera[taxa[5]].append(gid)
            gtdb_taxonomy[gid] = taxa
        
        # read NCBI metadata from GTDB metadata
        ncbi_species = {}
        ncbi_genus = {}
        ncbi_org_names = {}
        ncbi_isolate = {}
        wgs_id = {}
        with open(gtdb_metadata_file, 'rt') as f:
            header = f.readline().strip().split('\t')
            
            ncbi_taxonomy_index = header.index('ncbi_taxonomy')
            ncbi_org_name_index = header.index('ncbi_organism_name')
            ncbi_isolate_index = header.index('ncbi_isolate')
            wgs_acc_index = header.index('ncbi_wgs_master')
            
            for line in f:
                line_split = line.strip().split('\t')

                gid = line_split[0]

                ncbi_taxonomy = line_split[ncbi_taxonomy_index]
                if ncbi_taxonomy != 'none':
                    ncbi_species[gid] = [t.strip() for t in ncbi_taxonomy.split(';')][6]
                    ncbi_genus[gid] = [t.strip() for t in ncbi_taxonomy.split(';')][5]
                else:
                    ncbi_species[gid] = 's__'
                    ncbi_genus[gid] = 'g__'
                    
                ncbi_org_names[gid] = line_split[ncbi_org_name_index].replace('_', '-').upper()
                ncbi_isolate[gid] = line_split[ncbi_isolate_index].replace('_', '-').replace(' ', '-').upper()
                wgs_id[gid] = canonical_wgs_id(line_split[wgs_acc_index])

                if gid in uba_map:
                    ncbi_isolate[gid] = uba_map[gid]

        # validate generic species names in taxonomy file
        fout_pass = open(os.path.join(output_dir, 'supported_generic_names.tsv'), 'w')
        fout_pass.write('GTDB genus\tGenomes in genus\tGenomes supporting name\tReasons for support\n')
        
        fout_fail = open(os.path.join(output_dir, 'unsupported_generic_names.tsv'), 'w')
        fout_fail.write('GTDB genus\tGenomes in genus\tGenome IDs\n')
        
        for genus, gids in gtdb_genera.items():
            support = {}
            num_genomes = 0
            for gid in gids:
                if gid.startswith('D-'):
                    continue # skip dummy genomes
                    
                num_genomes += 1
                species = gtdb_taxonomy[gid][6]
                generic = species.split()[0].replace('s__', '')
                canonical_generic_name = canonical_taxon_name(generic)
                
                if generic != genus.replace('g__', ''):
                    self.logger.error('Generic species and genus names do not match: {} {} {}'.format(
                                        gid,
                                        species,
                                        genus))
                    sys.exit(-1)
                    
                # get valid genus names derived from NCBI genome accession
                cur_gid = gid
                if cur_gid in ncbi_map:
                    cur_gid = ncbi_map[gid]
                    
                genus_from_accession = []
                canonical_gid = cur_gid.replace('GB_', '').replace('RS_', '').replace('_', '-')
                canonical_gid = canonical_gid[0:canonical_gid.rfind('.')]
                genus_from_accession.append(canonical_gid)
                
                # accession without leading zeros
                t = canonical_gid[0:4]
                for idx, ch in enumerate(canonical_gid[4:]):
                    if ch != '0':
                        t += canonical_gid[4+idx:]
                        break
                    
                genus_from_accession.append(t)
                
                # check species ledger
                name_passed = False
                reason = None
                if gid in species_ledger:
                    ledger_species = species_ledger[gid]
                    ledger_generic = ledger_species.split()[0].replace('s__', '')
                    canonical_ledger_generic = canonical_taxon_name(ledger_generic)
                    if canonical_generic_name == canonical_ledger_generic:
                        support[gid] = 'Generic name specified in species ledger'
                elif gid in genus_ledger:
                    ledger_genus = genus_ledger[gid]
                    ledger_generic = ledger_genus.replace('g__', '')
                    canonical_ledger_generic = canonical_taxon_name(ledger_generic)
                    if canonical_generic_name == canonical_ledger_generic:
                        support[gid] = 'Generic name specified in genus ledger'
                elif (latinized_generic_sp_name(canonical_generic_name) 
                        and canonical_generic_name == ncbi_genus[gid].replace('g__', '')):
                    support[gid] = 'Congruent with NCBI genus name'
                elif canonical_generic_name.upper() in ncbi_org_names[gid]:
                    support[gid] = 'Generic name derived from NCBI organism name'
                elif canonical_generic_name.upper() in ncbi_isolate[gid]:
                    support[gid] = 'Generic name derived from NCBI infraspecific name'
                elif canonical_generic_name == wgs_id[gid]:
                    support[gid] = 'Generic name derived from NCBI WGS ID'
                elif canonical_generic_name in genus_from_accession:
                    support[gid] = 'Generic name derived from NCBI genome accession number'

            if len(support) > 0:
                support_str = []
                for k, v in support.items():
                    support_str.append('{}:{}'.format(k,v))
                fout_pass.write('{}\t{}\t{}\t{}\n'.format(
                                        genus, 
                                        num_genomes,
                                        len(support),
                                        '; '.join(support_str)))
            else:
                fout_fail.write('{}\t{}\t{}\n'.format(
                                genus, 
                                num_genomes, 
                                '; '.join(gids)))

        fout_fail.close()
        fout_pass.close()
        
    def check_specific(self, 
                    input_taxonomy,
                    gtdb_metadata_file,
                    sp_classification_ledger,
                    gtdb_synonym_file,
                    user_genome_id_table,
                    output_dir):
        """Validate specific name of species."""
        
        # read user genome ID table
        uba_map, ncbi_map = self.read_user_genome_table(user_genome_id_table)

        # read species ledger
        species_ledger = parse_species_ledgers(sp_classification_ledger, None)
        
        # read synonyms
        synonyms = parse_gtdb_synonym_file(gtdb_synonym_file)
                
        # read NCBI species from GTDB metadata
        ncbi_species = {}
        ncbi_subspecies = {}
        with open(gtdb_metadata_file, 'rt', encoding='utf-8') as f:
            header = f.readline().strip().split('\t')
            
            ncbi_taxonomy_index = header.index('ncbi_taxonomy')
            ncbi_unfiltered_index = header.index('ncbi_taxonomy_unfiltered')
            
            for line in f:
                line_split = line.strip().split('\t')

                gid = line_split[0]
                if gid in ncbi_map:
                    gid = ncbi_map[gid]
                elif gid in uba_map:
                    gid = uba_map[gid]
                    
                ncbi_taxonomy = line_split[ncbi_taxonomy_index]
                if ncbi_taxonomy != 'none':
                    ncbi_species[gid] = [t.strip() for t in ncbi_taxonomy.split(';')][6]

                ncbi_unfiltered = line_split[ncbi_unfiltered_index]
                if ncbi_unfiltered != 'none':
                    taxa = [t.strip() for t in ncbi_unfiltered.split(';')]
                    for taxon in taxa:
                        if taxon.startswith('sb__'):
                            ncbi_subspecies[gid] = taxon
                            break

        # validate specific name of genomes in taxonomy file
        fout_pass = open(os.path.join(output_dir, 'valid_specific_names.tsv'), 'w')
        fout_pass.write('Genome ID\tProposed species\tNCBI species\tReason for accepting\n')
        
        fout_fail = open(os.path.join(output_dir, 'invalid_specific_names.tsv'), 'w')
        fout_fail.write('Genome ID\tProposed species\tNCBI species\tReason\n')
        
        for line in open(input_taxonomy):
            line_split = line.strip().split('\t')
            
            gid = line_split[0]
            if gid.startswith('D-'):
                continue # skip dummy genomes
                
            if gid in ncbi_map:
                gid = ncbi_map[gid]
            elif gid in uba_map:
                gid = uba_map[gid]
                    
            species = line_split[1].split(';')[6].strip()
            generic_name, specific_name = species[3:].split()
            canonical_generic_name = canonical_taxon_name(generic_name)
            canonical_specific_name = canonical_taxon_name(specific_name)
            ncbi_sp = ncbi_species.get(gid, 's__')
            
            if (not latinized_generic_sp_name(canonical_generic_name) 
                and latinized_specific_sp_name(specific_name)):
                self.logger.error('Species names with placeholder generic names and Latinized specific names are invalid: {} {}'.format(
                                    gid, species))
            
            # check synonyms
            canonical_sp_name = canonical_species_name(species)
            if canonical_sp_name in synonyms:
                fout_fail.write('{}\t{}\t{}\t{}\n'.format(
                                    gid, 
                                    species, 
                                    ncbi_sp, 
                                    'Canonical species name is considered GTDB synonym of %s' % synonyms[canonical_sp_name].species))
                continue
            
            # check species ledger
            if gid in species_ledger:
                ledger_species = species_ledger[gid]
                ledger_specific = ledger_species.split()[1]
                canonical_ledger_specific = canonical_taxon_name(ledger_specific)
                if canonical_specific_name == canonical_ledger_specific:
                    fout_pass.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Specific name specified in species ledger'))
                else:
                    fout_fail.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Specific name conflicts with species ledger'))
                    
                continue

            if gtdb_canonical_species_id(gid, specific_name):
                # check for canonical GTDB identifier (i.e. sp<accession>)
                fout_pass.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Canonical GTDB species ID'))
            elif (latinized_specific_sp_name(canonical_specific_name) 
                    and canonical_specific_name in ncbi_sp):
                # check for agreement with NCBI
                fout_pass.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Congruent with NCBI specific name.'))
            elif (latinized_specific_sp_name(canonical_specific_name) 
                    and canonical_specific_name in ncbi_subspecies.get(gid, '')):
                # check for agreement with NCBI
                fout_pass.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Congruent with NCBI subspecies name.'))
            elif (latinized_specific_sp_name(canonical_specific_name) 
                    and canonical_specific_name not in ncbi_sp
                    and ncbi_sp != 's__'):
                # check for disagreement with NCBI
                fout_fail.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Incongruent with NCBI specific name.'))
            else:
                fout_fail.write('{}\t{}\t{}\t{}\n'.format(gid, species, ncbi_sp, 'Unknown source for specific name'))

        fout_fail.close()
        fout_pass.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import os
import sys
import logging
import re
import ast

import dendropy


class Taxa(object):
    def __init__(self, name, r=1):
        self.name = name
        self.reds = ['N/A'] * r
        self.distances = ['N/A'] * r
        self.medians = ['N/A'] * r


class RedDiff(object):
    """Class listing functions for RED analysis.

    """

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger()
        self.order_rank = ["d__", "p__", "c__", "o__", 'f__', 'g__', 's__']
        self.oneletter = {"domain": "d__",
                          "phylum": "p__",
                          "class": "c__",
                          "order": "o__",
                          "family": 'f__',
                          "genus": 'g__',
                          "species": 's__'}

    def reddiff(self, releases, output_file):
        outf = open(output_file, 'w')
        rank_dict = {k: {} for k in self.order_rank}
        number_or_releases = len(releases)
        versions = []

        header = ['Taxa']
        red_header = ['N/A'] * number_or_releases
        distance_header = ['N/A'] * number_or_releases
        median_header = ['N/A'] * number_or_releases
        for i, release in enumerate(releases):

            version, red_file, dict_file = release
            red_header[i] = 'RED value for release {}'.format(version)
            distance_header[i] = 'RED to median for release {}'.format(version)
            median_header[i] = 'rank median for release {}'.format(version)
            versions.append(version)
            dict_red = self.parse_dictfile(dict_file)

            with open(red_file, 'r') as rf:
                rf.readline()
                for line in rf:
                    taxa, taxonomy, red, meddiff = line.strip('\n').split('\t')[
                        0:4]
                    rk = taxa[0:3]
                    if taxa in rank_dict.get(rk):
                        taxtoupdate = rank_dict.get(rk).get(taxa)
                        taxtoupdate.reds[i] = red
                        taxtoupdate.distances[i] = meddiff
                        taxtoupdate.medians[i] = dict_red.get(rk)
                    else:
                        newtaxa = Taxa(taxa, number_or_releases)
                        newtaxa.reds[i] = red
                        newtaxa.distances[i] = meddiff
                        newtaxa.medians[i] = dict_red.get(rk)
                        rank_dict[rk][taxa] = newtaxa

        header.extend(red_header)
        header.extend(distance_header)
        header.extend(median_header)
        outf.write('{}\n'.format('\t'.join(header)))
        for rankletter in self.order_rank:
            if rankletter in rank_dict:
                for name, information in rank_dict.get(rankletter).items():
                    line_to_print = [name]
                    line_to_print.extend(information.reds)
                    line_to_print.extend(information.distances)
                    line_to_print.extend(information.medians)
                    outf.write('{}\n'.format(
                        '\t'.join([str(x) for x in line_to_print])))

    def parse_dictfile(self, dict_file):
        s = open(dict_file, 'r').read()
        temp_dict = ast.literal_eval(s)
        result = {self.oneletter.get(k): v for k, v in temp_dict.items()}
        return result



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import csv
import re
import logging
from collections import defaultdict

import dendropy

from fuzzywuzzy import fuzz

from biolib.common import check_file_exists, make_sure_path_exists, is_float
from biolib.external.execute import check_dependencies
from biolib.taxonomy import Taxonomy
from biolib.newick import parse_label

from gtdb_validation_tk.common import read_gtdb_metadata, canonical_species_name, canonical_wgs_id
from gtdb_validation_tk.check_taxonomy_file import CheckTaxonomyFile
from gtdb_validation_tk.suffix_table import SuffixTable
from gtdb_validation_tk.red_check import RED_Check
from gtdb_validation_tk.missing_genera import MissingGenera
from gtdb_validation_tk.merge_genera import MergeGenera
from gtdb_validation_tk.check_species_names import CheckSpeciesNames
from gtdb_validation_tk.validate_type_material import ValidateTypeMaterial
from gtdb_validation_tk.validate_species_by_ncbi_name import ValidateSpeciesNames
from gtdb_validation_tk.validate_gtdb_clusters_by_name import ValidateClustersNCBI
from gtdb_validation_tk.validate_species_clusters import ValidateSpeciesClusters
from gtdb_validation_tk.tree_diff import TreeDiff
from gtdb_validation_tk.tax_diff import TaxDiff
from gtdb_validation_tk.red_diff import RedDiff
from gtdb_validation_tk.spellchecker import SpellChecker

from numpy import mean as np_mean

csv.field_size_limit(min(2**31-1, sys.maxsize))


class OptionsParser(object):
    def __init__(self):
        """Initialization"""
        self.logger = logging.getLogger()

    def check_file(self, options):
        """Verify taxonomy file is formatted as expected."""

        check_file_exists(options.input_taxonomy)

        taxonomy = Taxonomy().read(options.input_taxonomy)
        
        # identify cases where root of child is highly similar to,
        # but not identical to root of parent
        # (e.g. f__Defferisomataceae; g__Deferrisoma -> missing 'f')
        self.logger.info('Identifying parent-child taxa with sufficiently similar, but non-identical names to suggest a type.')
        rank_suffix = {1: 'ota', 2: 'ia',  3: 'ales', 4: 'aceae', 5: None}
        potential_issues = {}
        for gid, taxa in taxonomy.items():
            for rank_idx in range(1, 5):
                # get parent name
                parent = taxa[rank_idx][3:]
                if '_' in parent:
                    parent = parent[0:parent.rfind('_')]
                    
                if not parent.endswith(rank_suffix[rank_idx]):
                    continue
                
                parent_root = parent.replace(rank_suffix[rank_idx], '')
                
                # get child name
                child = taxa[rank_idx+1][3:]
                if '_' in child:
                    child = child[0:child.rfind('_')]
                    
                if rank_suffix[rank_idx+1]:
                    if not child.endswith(rank_suffix[rank_idx+1]):
                        continue
                    child_root = child.replace(rank_suffix[rank_idx+1], '')
                else:
                    # genera have no fixed suffix
                    child_root = child
                    
                # limit roots to shortest string as trailing characters
                # often mismatch
                # (e.g.  f__Streptomycetaceae -> Streptomycet;
                #        g__Streptomyces -> Streptomyces)
                min_root_len = min(len(parent_root), len(child_root))
                parent_root = parent_root[0:min_root_len]
                child_root = child_root[0:min_root_len]
                
                fuzzy_score = fuzz.ratio(parent_root, child_root)
                prefix_score = 0
                for idx in range(min_root_len):
                    if parent_root[idx] != child_root[idx]:
                        break
                        
                    prefix_score += 1
                prefix_score = int(float(prefix_score)*100 / min_root_len)

                if (fuzzy_score != 100 and fuzzy_score >= 75 
                    and prefix_score != 100 and prefix_score <= 50 and prefix_score != 0):
                    potential_issues['{}#{}'.format(parent, child)] = (parent_root, child_root, fuzzy_score, prefix_score)

        print(' - identified {:,} potential issues'.format(len(potential_issues)))
        if len(potential_issues) > 0:
            for conflict, tokens in potential_issues.items():
                parent, child = conflict.split('#')
                parent_root, child_root, fuzzy_score, prefix_score = tokens
                print(' - {} {}: fuzzy score = {}, prefix score = {}'.format(
                        parent,
                        child,
                        fuzzy_score,
                        prefix_score))

        # perform standard taxonomy file checks
        p = CheckTaxonomyFile()
        p.check(taxonomy,
                check_prefixes=True,
                check_ranks=True,
                check_hierarchy=True,
                check_species=options.include_species,
                check_group_names=True,
                check_duplicate_names=True,
                check_capitalization=True)

        self.logger.info('Finished performing validation tests.')

    def check_tree(self, options):
        """Validate taxonomy of decorated tree and check for polyphyletic groups."""

        check_file_exists(options.decorated_tree)

        # validate taxonomy
        taxonomy = Taxonomy()
        if options.taxonomy_file:
            cur_taxonomy = taxonomy.read(options.taxonomy_file)
        else:
            cur_taxonomy = taxonomy.read_from_tree(options.decorated_tree)

        p = CheckTaxonomyFile()
        p.check(cur_taxonomy,
                check_prefixes=True,
                check_ranks=True,
                check_hierarchy=True,
                check_species=options.include_species,
                check_group_names=True,
                check_duplicate_names=True,
                check_capitalization=True)

        # check for polyphyletic groups
        polyphyletic_groups = set()
        tree = dendropy.Tree.get_from_path(options.decorated_tree,
                                           schema='newick',
                                           rooting="force-rooted",
                                           preserve_underscores=True)

        if options.taxonomy_file:
            # reduce taxonomy to taxa in tree and map taxon labels to Taxon
            # objects
            reduced_taxonomy = {}
            taxon_map = {}
            for leaf in tree.leaf_node_iter():
                if leaf.taxon.label in cur_taxonomy:  # need explicit check since tree may contain 'dummy' leaves
                    reduced_taxonomy[leaf.taxon.label] = cur_taxonomy[leaf.taxon.label]
                    taxon_map[leaf.taxon.label] = leaf.taxon

            # find taxa with an MRCA spanning additional taxa
            for rank_label in Taxonomy.rank_labels[1:]:
                extant_taxa = taxonomy.extant_taxa_for_rank(
                    rank_label, reduced_taxonomy)
                for taxon, taxa_ids in extant_taxa.items():
                    mrca = tree.mrca(taxa=[taxon_map[t] for t in taxa_ids])
                    mrca_leaf_count = sum(
                        [1 for leaf in mrca.leaf_iter() if leaf.taxon.label in cur_taxonomy])
                    if mrca_leaf_count != len(taxa_ids):
                        print(taxon, mrca_leaf_count, len(taxa_ids))
                        polyphyletic_groups.add(taxon)
        else:
            # find duplicate taxon labels in tree
            taxa = set()

            for node in tree.preorder_node_iter(lambda n: not n.is_leaf()):
                _support, taxon_label, _aux_info = parse_label(node.label)
                if taxon_label:
                    for taxon in [t.strip() for t in taxon_label.split(';')]:
                        if taxon in taxa:
                            polyphyletic_groups.add(taxon)

                        taxa.add(taxon)

        if len(polyphyletic_groups):
            print('')
            print('Tree contains polyphyletic groups:')
            for taxon in polyphyletic_groups:
                if not taxon.startswith('s__') or options.include_species:
                    print('%s' % taxon)
            print('')

        self.logger.info('Finished performing validation tests.')

    def check_suffix(self, options):
        """Report taxa with potentially erroneous polyphyly suffix designations."""

        check_file_exists(options.input_taxonomy)

        taxonomy = Taxonomy().read(options.input_taxonomy)

        # create set with all taxa in taxonomy
        all_taxa = set()
        for taxa in taxonomy.values():
            for taxon in taxa:
                all_taxa.add(taxon)

        # identify isolate names that may be erroneous
        suspecious_taxon_to_check = set()
        taxon_to_check = set()
        for gid, taxa in taxonomy.items():
            for taxon in taxa:
                if re.match('^[a-z][_]*[a-zA-Z0-9-]+[_]', taxon):
                    canonical_taxon = taxon[0:3] + ' '.join(
                        [t.strip() for t in re.split('_[A-Z]+(?= |$)', taxon[3:])]).strip()

                    if taxon.startswith('s__'):
                        if options.include_species:
                            genus = canonical_taxon.split(
                                ' ')[0].replace('s__', 'g__')
                            if canonical_taxon in all_taxa or genus in all_taxa:
                                # probably fine as canonical name exists in the
                                # taxonomy
                                taxon_to_check.add((taxon, canonical_taxon))
                            else:
                                suspecious_taxon_to_check.add(taxon)
                    else:
                        if canonical_taxon in all_taxa:
                            # probably fine as canonical name exists in the
                            # taxonomy
                            taxon_to_check.add((taxon, canonical_taxon))
                        else:
                            suspecious_taxon_to_check.add(taxon)

        if suspecious_taxon_to_check:
            print(
                '\nThe following taxa may have erroneous polyphyly suffix designations:')
            for taxon in suspecious_taxon_to_check:
                print('  %s' % taxon)

        if not taxon_to_check and not suspecious_taxon_to_check:
            print(
                '\nDid not identify any taxa with potentially erroneous polyphyly suffix designations.')

        self.logger.info('Finished performing test.')

    def red_check(self, options):
        """Report taxa with unexpected RED values."""

        check_file_exists(options.scaled_tree)
        check_file_exists(options.red_dictionary)

        p = RED_Check()
        p.run(options.scaled_tree,
              options.red_dictionary,
              options.output_file)

        self.logger.info('Results written to: %s' % options.output_file)
        self.logger.info('Finished performing validation tests.')

    def missing_genera(self, options):
        """Report lineages that may represent missing genus names."""

        check_file_exists(options.scaled_tree)
        check_file_exists(options.red_dictionary)
        check_file_exists(options.gtdb_prev_taxonomy)
        check_file_exists(options.ncbi_taxonomy)

        p = MissingGenera()
        p.run(options.scaled_tree,
              options.red_dictionary,
              options.gtdb_prev_taxonomy,
              options.ncbi_taxonomy,
              options.output_file)

        self.logger.info('Results written to: %s' % options.output_file)
        self.logger.info('Finished performing validation tests.')

    def merge_genera(self, options):
        """Report polyphyletic genera that are candidates for merging."""

        check_file_exists(options.scaled_tree)
        check_file_exists(options.red_dictionary)

        p = MergeGenera()
        p.run(options.scaled_tree,
              options.red_dictionary,
              options.output_file)

        self.logger.info('Results written to: %s' % options.output_file)
        self.logger.info('Finished performing validation tests.')

    def support(self, options):
        """Report taxa with low support values."""

        check_file_exists(options.decorated_tree)

        tree = dendropy.Tree.get_from_path(options.decorated_tree,
                                           schema='newick',
                                           rooting="force-rooted",
                                           preserve_underscores=True)

        poor_support = {}
        for n in tree.preorder_node_iter():
            support, taxon, _aux_info = parse_label(n.label)
            if not support:
                continue
                
            if taxon and support < options.min_support:
                poor_support[taxon] = support

        if poor_support:
            print('')
            for taxon in Taxonomy().sort_taxa(poor_support):
                print('%s\t%s' % (taxon, poor_support[taxon]))
            print('')

        self.logger.info('Finished performing test.')

    def tax_diff(self, options):
        """Compare two taxonomy files."""

        check_file_exists(options.input_taxonomy1)
        check_file_exists(options.input_taxonomy2)

        td = TaxDiff()
        td.tax_diff(options.input_taxonomy1,
                    options.input_taxonomy2,
                    options.gtdb_metadata_file,
                    options.output_dir)

        self.logger.info('Done.')

    def red_diff(self, options):
        """ Compare RED values between 2 releases """

        rd = RedDiff()
        rd.reddiff(options.release,
                   options.output_file)

        self.logger.info('Done.')

    def tree_diff(self, options):
        """Tree diff command."""

        check_file_exists(options.input_tree1)
        check_file_exists(options.input_tree2)
        make_sure_path_exists(options.output_dir)

        td = TreeDiff()
        td.run(options.input_tree1,
               options.input_tree2,
               options.output_dir,
               options.min_support,
               options.min_taxa)

        self.logger.info('Done.')

    def retired_taxa(self, options):
        """Report all taxa that have been retired (lost?) between two taxonomy files.."""

        check_file_exists(options.new_taxonomy_file)
        check_file_exists(options.prev_taxonomy_file)

        # get all taxa in previous taxonomy file
        prev_taxa = defaultdict(set)
        prev_gids = defaultdict(set)
        with open(options.prev_taxonomy_file) as f:
            for line in f:
                line_split = line.strip().split('\t')
                gid = line_split[0]
                taxa = [t.strip() for t in line_split[1].split(';')]
                for taxon in taxa:
                    rank_prefix = taxon[0:3]
                    prev_taxa[rank_prefix].add(taxon)
                    prev_gids[taxon].add(gid)

        # get all taxa in current taxonomy file
        cur_taxa = defaultdict(set)
        cur_taxonomy = {}
        with open(options.new_taxonomy_file) as f:
            for line in f:
                line_split = line.strip().split('\t')
                gid = line_split[0]
                taxa = [t.strip() for t in line_split[1].split(';')]
                cur_taxonomy[gid] = taxa
                for taxon in taxa:
                    rank_prefix = taxon[0:3]
                    cur_taxa[rank_prefix].add(taxon)

        fout = open(options.output_table, 'w')
        fout.write(
            'Rank\tRetired taxon\tClassification in new taxonomy\tGenome IDs in previous taxonomy\n')
        for idx, rank_prefix in enumerate(Taxonomy.rank_prefixes):
            for prev_taxon in prev_taxa[rank_prefix]:
                if prev_taxon not in cur_taxa[rank_prefix]:
                    new_classification = defaultdict(int)
                    for gid in prev_gids[prev_taxon]:
                        if gid in cur_taxonomy:
                            new_classification[cur_taxonomy[gid][idx]] += 1

                    new_classification_str = []
                    for taxon, count in new_classification.items():
                        new_classification_str.append(
                            '{}:{}'.format(taxon, count))

                    fout.write('{}\t{}\t{}\t{}\n'.format(
                        Taxonomy.rank_labels[idx],
                        prev_taxon,
                        ', '.join(new_classification_str),
                        ', '.join(prev_gids[prev_taxon])))

        fout.close()

        self.logger.info('Done.')

    def branch_len(self, options):
        """Branch length command."""

        check_file_exists(options.input_tree)

        tree = dendropy.Tree.get_from_path(options.input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        leaf_root_dists = tree.calc_node_root_distances(
            return_leaf_distances_only=True)
        mean_branch_len = np_mean(leaf_root_dists)

        self.logger.info(
            'Mean root to extant taxa branch length is %.3f.' % mean_branch_len)

        if mean_branch_len < 0.5 or mean_branch_len > 2.0:
            self.logger.warning(
                'Mean root to extant taxa branch length is unusual.')
            self.logger.warning(
                'This value depend on the rooting of the tree, but is expected to be between 0.5 and 2.0.')
        else:
            self.logger.info('This is within the range expected, [0.5, 2.0].')

        self.logger.info('Done.')

    def check_spelling(self, options):
        """Return a Levenshtein distance between ranks in 2 different taxonomy. This can be used to detect typos."""

        check_file_exists(options.input_taxonomy1)
        check_file_exists(options.input_taxonomy2)
        sc = SpellChecker()
        sc.spell_check(options.input_taxonomy1,
                       options.input_taxonomy2,
                       options.dist,
                       options.output_file)
        self.logger.info('Done.')

    def check_distrib(self, options):
        """Return summary of group changes between 2 taxonomies."""

        check_file_exists(options.input_taxonomy1)
        check_file_exists(options.input_taxonomy2)
        make_sure_path_exists(options.output_dir)

        td = TaxDiff()
        td.check_distrib(options.input_taxonomy1,
                         options.input_taxonomy2,
                         options.output_dir,
                         options.keep_no_change,
                         options.keep_placeholder_name)
        self.logger.info('Done.')

    def ncbi_tax_diff(self, options):
        """Establish changes in NCBI taxonomy that may impact GTDB taxonomy."""

        check_file_exists(options.prev_ncbi_taxonomy)
        check_file_exists(options.cur_ncbi_taxonomy)
        check_file_exists(options.gtdb_taxonomy)

        td = TaxDiff()
        td.ncbi_tax_diff(options.prev_ncbi_taxonomy,
                         options.cur_ncbi_taxonomy,
                         options.gtdb_taxonomy,
                         options.output_dir)

        self.logger.info('Done.')

    def check_generic(self, options):
        """Validate generic name of species."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.sp_classification_ledger)
        check_file_exists(options.genus_ledger)
        check_file_exists(options.gtdb_synonym_file)
        check_file_exists(options.user_genome_id_table)
        make_sure_path_exists(options.output_dir)

        p = CheckSpeciesNames()
        p.check_generic(options.input_taxonomy,
                        options.gtdb_metadata_file,
                        options.sp_classification_ledger,
                        options.genus_ledger,
                        options.gtdb_synonym_file,
                        options.user_genome_id_table,
                        options.output_dir)

        self.logger.info('Done.')

    def check_specific(self, options):
        """Validate specific name of species."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.sp_classification_ledger)
        check_file_exists(options.gtdb_synonym_file)
        check_file_exists(options.user_genome_id_table)
        make_sure_path_exists(options.output_dir)

        p = CheckSpeciesNames()
        p.check_specific(options.input_taxonomy,
                         options.gtdb_metadata_file,
                         options.sp_classification_ledger,
                         options.gtdb_synonym_file,
                         options.user_genome_id_table,
                         options.output_dir)

        self.logger.info('Done.')

    def species_clusters(self, options):
        """Compare GTDB taxonomy to type strains."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        make_sure_path_exists(options.output_dir)

        p = ValidateSpeciesClusters()
        p.run(options.input_taxonomy,
              options.gtdb_metadata_file,
              options.output_dir)

        self.logger.info('Done.')

    def type_strains(self, options):
        """Compare GTDB taxonomy to type strains."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.sp_classification_ledger)
        check_file_exists(options.gtdb_named_type_genomes)
        check_file_exists(options.gtdb_synonym_file)
        make_sure_path_exists(options.output_dir)

        v = ValidateTypeMaterial()
        v.validate_type_strains(options.input_taxonomy,
                                options.gtdb_metadata_file,
                                options.sp_classification_ledger,
                                options.gtdb_named_type_genomes,
                                options.gtdb_synonym_file,
                                options.show_synonyms,
                                options.prefix,
                                options.output_dir)

        self.logger.info('Done.')

    def type_species(self, options):
        """Compare GTDB taxonomy to type species."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.gtdb_synonym_file)
        check_file_exists(options.sp_classification_ledger)
        check_file_exists(options.genus_classification_ledger)
        make_sure_path_exists(options.output_dir)

        v = ValidateTypeMaterial()
        v.validate_type_species(options.input_taxonomy,
                                options.gtdb_metadata_file,
                                options.gtdb_synonym_file,
                                options.sp_classification_ledger,
                                options.genus_classification_ledger,
                                options.show_synonyms,
                                options.prefix,
                                options.output_dir)

        self.logger.info('Done.')

    def val_table(self, options):
        """General validation table."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)

        gtdb_taxonomy = {}
        with open(options.input_taxonomy) as f:
            for line in f:
                line_split = line.strip().split('\t')
                gid = line_split[0]
                if gid.startswith('D-'):
                    continue  # dummy genome_dir_file

                taxa = [t.strip() for t in line_split[1].split(';')]
                gtdb_taxonomy[gid] = taxa

        fout = open(options.output_table, 'w')
        fout.write('Genome ID\tProposed taxonomy\tProposed species')
        fout.write(
            '\tNCBI taxonomy\tNCBI species\tNCBI organism name\tNCBI infraspecific name\tNCBI WGS ID')
        fout.write('\tGTDB type designation\n')
        with open(options.gtdb_metadata_file) as f:
            header = f.readline().strip().split('\t')

            ncbi_taxonomy_index = header.index('ncbi_taxonomy')
            ncbi_org_name_index = header.index('ncbi_organism_name')
            ncbi_wgs_index = header.index('ncbi_wgs_master')
            ncbi_isolate_index = header.index('ncbi_isolate')
            gtdb_type_designation_index = header.index('gtdb_type_designation')

            for line in f:
                line_split = line.strip().split('\t')

                gid = line_split[0]
                if gid not in gtdb_taxonomy:
                    continue

                ncbi_taxonomy = line_split[ncbi_taxonomy_index]
                ncbi_taxa = Taxonomy.rank_prefixes
                if ncbi_taxonomy and ncbi_taxonomy != 'none':
                    ncbi_taxa = [t.strip() for t in ncbi_taxonomy.split(';')]

                ncbi_org_name = line_split[ncbi_org_name_index]
                if not ncbi_org_name or ncbi_org_name == 'none':
                    ncbi_org_name = 'n/a'

                ncbi_isolate = line_split[ncbi_isolate_index]
                if not ncbi_isolate or ncbi_isolate == 'none':
                    ncbi_isolate = 'n/a'

                ncbi_wgs_id = canonical_wgs_id(line_split[ncbi_wgs_index])

                gtdb_type_designation = line_split[gtdb_type_designation_index]

                fout.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(
                    gid,
                    '; '.join(gtdb_taxonomy[gid]),
                    gtdb_taxonomy[gid][6],
                    '; '.join(ncbi_taxa),
                    ncbi_taxa[6],
                    ncbi_org_name,
                    ncbi_isolate,
                    ncbi_wgs_id,
                    gtdb_type_designation))

        fout.close()

    def ncbi_species(self, options):
        """Report differences between GTDB and NCBI species names."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        make_sure_path_exists(options.output_dir)

        validate = ValidateSpeciesNames()
        validate.run(options.input_taxonomy,
                     options.gtdb_metadata_file,
                     options.output_dir)

        self.logger.info('Done.')

    def gtdb_cluster_ncbi(self, options):
        """Report NCBI species within each GTDB-defined cluster."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.gtdb_metadata_file)
        make_sure_path_exists(options.output_dir)

        validate = ValidateClustersNCBI()
        validate.run(options.input_taxonomy,
                     options.gtdb_metadata_file,
                     options.output_dir)

        self.logger.info('Done.')

    def suffix_table(self, options):
        """Create table indicating last polyphyly suffix used for each taxon."""

        s = SuffixTable()
        s.run(options.input_taxonomy, options.output_table)

        self.logger.info('Done.')

    def add_dummy(self, options):
        """Add a zero branch length sister node to each current leaf node."""

        # add dummy nodes to tree
        check_file_exists(options.input_tree)

        self.logger.info('Reading input tree.')
        tree = dendropy.Tree.get_from_path(options.input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)
        self.logger.info(' ...identified {} leaf nodes.'.format(
            sum([1 for leaf in tree.leaf_node_iter()])))

        # create zero branch length sister node to each leaf node
        self.logger.info('Adding dummy leaf nodes to tree.')
        for leaf in tree.leaf_node_iter():
            gid = leaf.taxon.label

            leaf.label = None
            leaf.taxon = None
            leaf.new_child(label=gid,
                           edge_length=0.0)
            leaf.new_child(label='D-' + gid,
                           edge_length=0.0)

        # write new tree to file
        base, ext = os.path.splitext(options.input_tree)
        output_tree = base + '.dummy_nodes' + ext
        self.logger.info(
            'Writing modified tree to: {}'.format(output_tree))
        tree.write_to_path(output_tree,
                           schema='newick',
                           suppress_rooting=True,
                           suppress_leaf_node_labels=False,
                           unquoted_underscores=True)

        # add dummy nodes to taxonomy file
        if options.taxonomy_file:
            self.logger.info('Adding dummy nodes to taxonomy file.')
            
            base, ext = os.path.splitext(options.taxonomy_file)
            output_taxonomy_file = base + '.dummy_nodes' + ext
            fout = open(output_taxonomy_file, 'w')
            
            with open(options.taxonomy_file) as f:
                for line in f:
                    fout.write(line)
                    fout.write('D-'+line)
          
            fout.close()
            
            self.logger.info(
            'Writing modified taxonomy file to: {}'.format(output_taxonomy_file))

        self.logger.info('Done.')

    def remove_dummy(self, options):
        """Remove dummy nodes from curation tree."""

        check_file_exists(options.input_tree)

        self.logger.info('Reading input tree.')
        tree = dendropy.Tree.get_from_path(options.input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)
        self.logger.info(' ...identified {} leaf nodes.'.format(
            sum([1 for leaf in tree.leaf_node_iter()])))

        # find dummy nodes in tree
        self.logger.info('Identifying dummy nodes to prune.')
        dummy_nodes = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label.startswith('D-'):
                dummy_nodes.add(leaf.taxon)

        self.logger.info(
            'Identified %d dummy nodes to prune from tree.' % len(dummy_nodes))

        # prune tree
        self.logger.info('Pruning tree.')
        tree.prune_taxa(dummy_nodes)

        # write new tree to file
        self.logger.info(
            'Writing modified tree to: {}'.format(options.output_tree))
        tree.write_to_path(options.output_tree,
                           schema='newick',
                           suppress_rooting=True,
                           suppress_leaf_node_labels=False,
                           unquoted_underscores=True)

        self.logger.info('Done.')

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)"""

        logging.basicConfig(format='', level=logging.INFO)
        if options.subparser_name == 'check_file':
            self.check_file(options)
        elif options.subparser_name == 'check_tree':
            self.check_tree(options)
        elif options.subparser_name == 'check_suffix':
            self.check_suffix(options)
        elif options.subparser_name == 'check_spelling':
            self.check_spelling(options)
        elif options.subparser_name == 'check_distrib':
            self.check_distrib(options)
        elif options.subparser_name == 'red_check':
            self.red_check(options)
        elif options.subparser_name == 'missing_genera':
            self.missing_genera(options)
        elif options.subparser_name == 'merge_genera':
            self.merge_genera(options)
        elif options.subparser_name == 'support':
            self.support(options)
        elif options.subparser_name == 'red_diff':
            self.red_diff(options)
        elif options.subparser_name == 'tax_diff':
            self.tax_diff(options)
        elif options.subparser_name == 'tree_diff':
            self.tree_diff(options)
        elif options.subparser_name == 'retired_taxa':
            self.retired_taxa(options)
        elif options.subparser_name == 'branch_len':
            self.branch_len(options)
        elif options.subparser_name == 'ncbi_tax_diff':
            self.ncbi_tax_diff(options)
        elif options.subparser_name == 'check_generic':
            self.check_generic(options)
        elif options.subparser_name == 'check_specific':
            self.check_specific(options)
        elif options.subparser_name == 'species_clusters':
            self.species_clusters(options)
        elif options.subparser_name == 'type_strains':
            self.type_strains(options)
        elif options.subparser_name == 'type_species':
            self.type_species(options)
        elif options.subparser_name == 'val_table':
            self.val_table(options)
        elif options.subparser_name == 'ncbi_species':
            self.ncbi_species(options)
        elif options.subparser_name == 'gtdb_cluster_ncbi':
            self.gtdb_cluster_ncbi(options)
        elif options.subparser_name == 'suffix_table':
            self.suffix_table(options)
        elif options.subparser_name == 'add_dummy':
            self.add_dummy(options)
        elif options.subparser_name == 'remove_dummy':
            self.remove_dummy(options)
        else:
            self.logger.error('Unknown command: ' +
                              options.subparser_name + '\n')
            sys.exit()

        return 0



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import os
import sys
import logging
import re

from collections import defaultdict
from collections import Counter

from biolib.newick import parse_label
from biolib.taxonomy import Taxonomy
import dendropy


class TaxDiff(object):
    """Tabulate differences between two taxonomies.

    Every taxon label in the first taxonomy is classified
    to indicate how it has changed relative to the
    second taxonomy.

    This implies that every label in the first taxonomy is
    given precisely one label and that labels in the second 
    taxonomy may often be seen simply as deprecated. 

    Taxon are classified as follows in the order listed:

    retained: the exact same taxon label is in the GTDB and NCBI taxonomies
    corrected: only the suffix of the taxon has changed
    more specific: taxon has been moved down 1 or more ranks (e.g., phylum to class)
    more general: taxon has been moved up 1 or more ranks (e.g., class to phylum)
    new: the label has been introduced by the GTDB taxonomy
    deprecated: the label has been removed in the GTDB taxonomy

    E.g.,
    The first taxonomy has the label o__Archaeoglobales,
    but no label c__Archaeoglobi. The second taxonomy has
    the labels c__Archaeoglobi and o__Archaeoglobales. AssertionError
    such, o__Archaeoglobales will be classified as retained
    while c__Archaeoglobi will be classified as deprecated
    (and not 'more specific').
    """

    def __init__(self, dpi=96):
        """Initialize."""
        self.logger = logging.getLogger()

        # c__Halobacteria => o__Halobacteriales
        # c__Hadesarchaea => c__Hadesarchaeia
        # c__Archaeoglobi => o__Archaeoglobales

        # p__Pacearchaeota => f__Pacearchaeaceae

        # domain = 'a', phylum = 'aeota', class = 'ia', order = 'ales', family
        # = 'aceae'
        self.rank_suffices = ['a', 'aeota', 'ia', 'ales', 'aceae', '']

    def _renamed(self, taxon, cur_rank, taxa_at_rank):
        """Determine if taxon has been renamed."""

        # remove paraphyletic suffix if one exists
        taxon = taxon[0:3] + taxon[3:].split('_')[0]

        # determine order of ranks to examine in order
        # to first identify 'more specific' and then
        # 'more general' label changes
        rank_order = [r for r in range(cur_rank - 1, -1, -1)]
        rank_order += [r for r in range(cur_rank + 1,
                                        len(Taxonomy.rank_prefixes))]
        # check if taxon has been moved up or down in rank
        # as determined by removing suffix characters and
        # appending canonical rank prefixes
        for old_rank in rank_order:
            if old_rank >= len(self.rank_suffices):
                return None, None

            # eat suffix character by character and append canonical suffix
            old_rank_prefix = Taxonomy.rank_prefixes[old_rank]
            cur_suffix = self.rank_suffices[cur_rank]
            old_suffix = self.rank_suffices[old_rank]
            for i in range(1, max([len(s) for s in self.rank_suffices]) + 1):  # eat taxon name

                if i == 0:
                    modified_taxon = taxon[3:]
                else:
                    modified_taxon = taxon[3:-i]

                for j in range(0, len(old_suffix)):  # eat suffix
                    old_taxon = old_rank_prefix + \
                        modified_taxon + old_suffix[j:]
                    if old_taxon in taxa_at_rank[old_rank]:
                        return old_taxon, old_rank

                    old_taxon = old_rank_prefix + \
                        modified_taxon + old_suffix[0:j]
                    if old_taxon in taxa_at_rank[old_rank]:
                        return old_taxon, old_rank

        return None, None

    def _change_suffix(self, taxon, cur_rank, taxa_at_rank):
        """Change suffix of taxon."""

        # check if taxon name has been corrected
        for old_rank, old_suffix in enumerate(self.rank_suffices):
            # eat suffix character by character and append canonical suffix
            for i in range(0, max([len(s) for s in self.rank_suffices])):  # eat taxon name
                if i == 0:
                    modified_taxon = taxon
                else:
                    modified_taxon = taxon[0:-i]

                for j in range(0, len(old_suffix)):  # eat suffix
                    old_taxon = modified_taxon + old_suffix[j:]
                    if old_taxon in taxa_at_rank[cur_rank]:
                        return old_taxon

                    old_taxon = modified_taxon + old_suffix[0:j]
                    if old_taxon in taxa_at_rank[cur_rank]:
                        return old_taxon

        return None

    def _tax_diff_table(self, tax1, tax2, output_table):
        """Tabulate incongruency of taxonomy strings at each rank."""

        fout = open(output_table, 'w')
        fout.write('Lineage\tNo. Extent Taxa')
        for rank_label in Taxonomy.rank_labels:
            fout.write('\t%s (%%)' % rank_label.title())
        fout.write('\n')

        taxonomy = Taxonomy()
        named_lineages_at_rank = taxonomy.named_lineages_at_rank(tax1)
        for rank, taxa in named_lineages_at_rank.items():
            rank_label = Taxonomy.rank_labels[rank]
            if rank_label == 'species':
                continue

            extant_taxa_for_rank = taxonomy.extant_taxa_for_rank(
                rank_label, tax1)

            for taxon in taxa:
                extent_taxa = extant_taxa_for_rank[taxon]
                fout.write('%s\t%d' % (taxon, len(extent_taxa)))

                row = defaultdict(list)
                for genome_id in extent_taxa:
                    taxa1 = tax1[genome_id]
                    taxa2 = tax2[genome_id]

                    for cur_rank, (taxa1, taxa2) in enumerate(zip(taxa1, taxa2)):
                        row[cur_rank].append(taxa1 == taxa2)

                for cur_rank, matches in row.items():
                    if cur_rank <= rank:
                        fout.write('\t-')
                    else:
                        perc_match = sum(matches) * 100.0 / len(matches)
                        fout.write('\t%.1f' % (100.0 - perc_match))
                fout.write('\n')
        fout.close()

    def tree_tax_diff(self, tree1_file, tree2_file, output_dir):
        """Tabulate differences between two taxonomies on a tree.

        Parameters
        ----------
        tree1_file : str
            File with tree in Newick format.
        tree2_file : str
            File with tree in Newick format.
        output_dir : str
            Output directory.
        """

        tree1 = dendropy.Tree.get_from_path(tree1_file,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        tree2 = dendropy.Tree.get_from_path(tree2_file,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        # prune both trees to a set of common taxa
        taxa1 = set()
        for t in tree1.leaf_node_iter():
            taxa1.add(t.taxon.label)

        taxa2 = set()
        for t in tree2.leaf_node_iter():
            taxa2.add(t.taxon.label)

        taxa_in_common = taxa1.intersection(taxa2)
        self.logger.info('Tree 1 contains %d taxa.' % len(taxa1))
        self.logger.info('Tree 2 contains %d taxa.' % len(taxa2))
        self.logger.info('Pruning trees to the %d taxa in common.' %
                         len(taxa_in_common))

        tree1.retain_taxa_with_labels(taxa_in_common)
        tree2.retain_taxa_with_labels(taxa_in_common)

        # get named lineages at each taxonomic rank
        taxonomy = Taxonomy()
        tax1 = taxonomy.read_from_tree(tree1)
        tax2 = taxonomy.read_from_tree(tree2)

        taxa_at_rank1 = taxonomy.named_lineages_at_rank(tax1)
        taxa_at_rank2 = taxonomy.named_lineages_at_rank(tax2)

        # identify retained taxonomic names
        tax_file_name = os.path.splitext(os.path.basename(tree1_file))[0]
        output_file = os.path.join(
            output_dir, '%s.taxa_diff.tsv' % tax_file_name)
        fout = open(output_file, 'w')
        fout.write('Rank\tClassification\tTaxonomy 1\tTaxonomy 2\n')
        taxon2_accounted_for = defaultdict(set)
        for rank, rank_label in enumerate(Taxonomy.rank_labels[0:-1]):
            for taxon in taxa_at_rank1[rank]:
                # check if taxon has been retained
                if taxon in taxa_at_rank2[rank]:
                    fout.write('%s\t%s\t%s\t%s\n' %
                               (rank_label, 'retained', taxon, taxon))
                    taxon2_accounted_for[rank].add(taxon)
                    continue

                # check if name was simply corrected by changing suffix
                old_taxon = self._change_suffix(taxon, rank, taxa_at_rank2)
                if old_taxon:
                    fout.write('%s\t%s\t%s\t%s\n' %
                               (rank_label, 'corrected', taxon, old_taxon))
                    taxon2_accounted_for[rank].add(old_taxon)
                    continue

                # check if taxon has been moved up or down in rank
                old_taxon, old_rank = self._renamed(taxon, rank, taxa_at_rank2)
                if old_taxon:
                    if rank < old_rank:
                        fout.write('%s\t%s\t%s\t%s\n' % (
                            rank_label, 'more general', taxon, old_taxon))
                    elif rank == old_rank:
                        fout.write('%s\t%s\t%s\t%s\n' %
                                   (rank_label, 'corrected', taxon, old_taxon))
                    else:
                        fout.write('%s\t%s\t%s\t%s\n' % (
                            rank_label, 'more specific', taxon, old_taxon))

                    taxon2_accounted_for[old_rank].add(old_taxon)
                    continue

                # otherwise, the taxon appears to be new
                fout.write('%s\t%s\t%s\t%s\n' %
                           (rank_label, 'new', taxon, 'NA'))

        # report deprecated taxa
        for rank, rank_label in enumerate(Taxonomy.rank_labels[0:-1]):
            for taxon in taxa_at_rank2[rank]:
                if taxon not in taxon2_accounted_for[rank]:
                    fout.write('%s\t%s\t%s\t%s\n' %
                               (rank_label, 'deprecated', 'NA', taxon))

        fout.close()

        # tabulate congruence of taxonomy strings
        output_table = os.path.join(
            output_dir, '%s.perc_diff.tsv' % tax_file_name)
        self._tax_diff_table(tax1, tax2, output_table)

    def _read_type_material(self, gtdb_metadata_file):
        """Read type material metadata."""

        metadata = {}
        with open(gtdb_metadata_file) as f:
            header = f.readline().strip().split('\t')

            gtdb_type_material_index = header.index('gtdb_type_material')
            gtdb_type_material_sources_index = header.index(
                'gtdb_type_material_sources')

            for line in f:
                line_split = line.strip().split('\t')

                gid = line_split[0]

                gtdb_type_material = line_split[gtdb_type_material_index]
                gtdb_type_material_sources = line_split[gtdb_type_material_sources_index]

                metadata[gid] = [gtdb_type_material,
                                 gtdb_type_material_sources]

        return metadata

    def tax_diff(self, tax1_file, tax2_file, gtdb_metadata_file, output_dir):
        """Tabulate differences between two taxonomies.

        Parameters
        ----------
        tax1_file : str
            First taxonomy file.
        tax2_file : str
            Second taxonomy file.
        output_dir : str
            Output directory.
        """

        gtdb_metadata = None
        if gtdb_metadata_file:
            gtdb_metadata = self._read_type_material(gtdb_metadata_file)

        tax1 = Taxonomy().read(tax1_file)
        tax2 = Taxonomy().read(tax2_file)
        common_taxa = set(tax1.keys()).intersection(tax2.keys())

        self.logger.info('First taxonomy contains %d taxa.' % len(tax1))
        self.logger.info('Second taxonomy contains %d taxa.' % len(tax2))
        self.logger.info('Taxonomies have %d taxa in common.' %
                         len(common_taxa))

        # identify differences between taxonomies
        tax_file_name1 = os.path.splitext(os.path.basename(tax1_file))[0]
        tax_file_name2 = os.path.splitext(os.path.basename(tax2_file))[0]
        output_table = os.path.join(output_dir, 'classification_table.tsv')

        fout = open(output_table, 'w')
        fout.write('Genome ID\tChange\tRank\t%s\t%s' %
                   (tax_file_name1, tax_file_name2))
        if gtdb_metadata:
            fout.write('\tType material\tType material sources')
        fout.write('\n')

        # T2 = g__Bob -> T1 = g__Bob, or T2 = g__ -> T1 = g__
        unchanged = defaultdict(int)
        # T2 = g__Bob -> T1 = g__Jane, or T2 = g__Bob -> T1 = g__Bob_A
        active_change = defaultdict(int)
        passive_change = defaultdict(int)      # T2 = g__??? -> T1 = g__Jane
        unresolved_change = defaultdict(int)   # T2 = g__Box -> T1 = g__???
        for taxa in common_taxa:
            t1 = tax1[taxa]
            t2 = tax2[taxa]

            for rank, (taxon1, taxon2) in enumerate(zip(t1, t2)):
                bChange = False
                if taxon1 == taxon2:
                    unchanged[rank] += 1
                elif taxon1 != Taxonomy.rank_prefixes[rank] and taxon2 != Taxonomy.rank_prefixes[rank]:
                    active_change[rank] += 1
                    fout.write('%s\t%s\t%s\t%s\t%s' % (
                        taxa, 'active', Taxonomy.rank_labels[rank], ';'.join(t1), ';'.join(t2)))
                    bChange = True
                elif taxon2 == Taxonomy.rank_prefixes[rank]:
                    passive_change[rank] += 1
                    fout.write('%s\t%s\t%s\t%s\t%s' % (
                        taxa, 'passive', Taxonomy.rank_labels[rank], ';'.join(t1), ';'.join(t2)))
                    bChange = True
                elif taxon1 == Taxonomy.rank_prefixes[rank]:
                    unresolved_change[rank] += 1
                    fout.write('%s\t%s\t%s\t%s\t%s' % (
                        taxa, 'unresolved', Taxonomy.rank_labels[rank], ';'.join(t1), ';'.join(t2)))
                    bChange = True

                if bChange:
                    if gtdb_metadata:
                        fout.write('\t%s\t%s' % (
                            gtdb_metadata[taxa][0], gtdb_metadata[taxa][1]))
                    fout.write('\n')

        fout.close()

        # report results
        output_table = os.path.join(
            output_dir, 'classification_summary_table.tsv')

        fout = open(output_table, 'w')
        fout.write(
            'Rank\tUnchanged\tUnchanged (%)\tActive\t Active (%)\tPassive\tPassive (%)\tUnresolved\tUnresolved (%)\n')
        print('')
        print('Rank\tUnchanged\tActive\tPassive\tUnresolved\tTotal')
        for rank in range(0, len(Taxonomy.rank_prefixes)):
            total = unchanged[rank] + active_change[rank] + \
                passive_change[rank] + unresolved_change[rank]
            if total != 0:
                fout.write('%s\t%d\t%.1f\t%d\t%.1f\t%d\t%.1f\t%d\t%.1f\n' %
                           (Taxonomy.rank_labels[rank],
                            unchanged[rank], unchanged[rank] * 100.0 / total,
                            active_change[rank], active_change[rank] *
                            100.0 / total,
                            passive_change[rank], passive_change[rank] *
                            100.0 / total,
                            unresolved_change[rank], unresolved_change[rank] * 100.0 / total))
                print('%s\t%d\t%d\t%d\t%d\t%d' % (Taxonomy.rank_labels[rank],
                                                  unchanged[rank],
                                                  active_change[rank],
                                                  passive_change[rank],
                                                  unresolved_change[rank],
                                                  total))
        print('')

    def ncbi_tax_diff(self,
                      prev_ncbi_tax_file,
                      cur_ncbi_tax_file,
                      gtdb_tax_file,
                      gtdb_metadata_file,
                      output_dir):
        """Establish changes in NCBI taxonomy that may impact GTDB taxonomy.

        Parameters
        ----------
        prev_ncbi_tax_file : str
            Previous NCBI taxonomy file.
        cur_ncbi_tax_file : str
            Current NCBI taxonomy file.
        gtdb_tax_file
            GTDB taxonomy file.
        output_dir : str
            Output directory.
        """

        # get type material metadata
        type_material = {}
        with open(gtdb_metadata_file) as f:
            header = f.readline().strip().split('\t')

            gtdb_type_material_index = header.index('gtdb_type_material')
            gtdb_type_material_sources_index = header.index(
                'gtdb_type_material_sources')

            for line in f:
                line_split = line.strip().split('\t')

                gid = line_split[0]
                gtdb_type_material = line_split[gtdb_type_material_index]
                if gtdb_type_material.lower() == 't' or gtdb_type_material.lower() == 'true':
                    gtdb_type_material_sources = line_split[gtdb_type_material_sources_index]
                    type_material[gid] = gtdb_type_material_sources

        # identify differences in NCBI taxonomies
        prev_ncbi = Taxonomy().read(prev_ncbi_tax_file)
        cur_ncbi = Taxonomy().read(cur_ncbi_tax_file)
        gtdb = Taxonomy().read(gtdb_tax_file)

        fout = open(os.path.join(output_dir, 'ncbi_tax_diff.tsv'), 'w')
        fout.write(
            'Genome ID\tPrevious NCBI taxonomy\tCurrent NCBI taxonomy\tHighest modified rank')
        fout.write('\tGTDB taxonomy\tImpact GTDB taxonomy\tGTDB type material\n')
        for gid in set(prev_ncbi).intersection(cur_ncbi):
            for r, rank_label in enumerate(Taxonomy.rank_labels):
                prev_taxon = prev_ncbi[gid][r]
                cur_taxon = cur_ncbi[gid][r]
                if cur_taxon != prev_taxon:
                    fout.write('%s\t%s\t%s\t%s' % (gid,
                                                   '; '.join(prev_ncbi[gid]),
                                                   '; '.join(cur_ncbi[gid]),
                                                   rank_label))

                    gtdb_id = gid
                    if gtdb_id.startswith('GCA_'):
                        gtdb_id = 'GB_' + gtdb_id
                    elif gtdb_id.startswith('GCF_'):
                        gtdb_id = 'RS_' + gtdb_id

                    if gtdb_id in gtdb:
                        if gtdb[gtdb_id][r] == prev_taxon:
                            fout.write('\t%s\tTRUE' % '; '.join(gtdb[gtdb_id]))
                        else:
                            fout.write('\t%s\tFALSE' %
                                       '; '.join(gtdb[gtdb_id]))
                    else:
                        fout.write('\t-\tFALSE')

                    fout.write('\t%s' % type_material.get(gtdb_id, '-'))
                    fout.write('\n')
                    break

    def check_distrib(self, old_taxonomy, new_taxonomy, output_directory, keep_no_change, keep_placeholder_name, top_change):

        gtogfile = open(os.path.join(
            output_directory, 'genomes_summary.tsv'), "w")
        gtogfile.write("genome_id\tfirst_taxonomy\tsecond_taxonomy\tstatus\n")
        for dom in ["d__Archaea", "d__Bacteria"]:
            tempdict = {}  # temporary dictionary for the information in old taxonomy
            olddict = {}
            newdict = {}
            mapdict = {}  # dictionary of uba_mappings
            oldnumber = 0  # number of genes in old_tax
            newnumber = 0  # number of genes in new_tax
            intersection = 0  # number of genes common to both taxonomies

            with open("uba_mapping/uba_mapping_v2.tsv", "r") as tax_map:
                for line in tax_map:
                    genes_id = line.strip().split("\t")
                    # IDs that start with U_
                    u_id = genes_id[0]
                    # IDs that start with GCA or GCF
                    gc_id = genes_id[2]
                    # delete GB or RS prefix to ensure id-consistency
                    if gc_id.startswith("GB") or gc_id.startswith("RS"):
                        gc_id = gc_id[3:]
                    # map k, v of all id_mappings
                    mapdict[gc_id] = u_id
                    mapdict[u_id] = gc_id

            with open(new_taxonomy) as taxnew:
                for line in taxnew:
                    newnumber += 1
                    genome_info = line.strip().split("\t")
                    gene_id = genome_info[0]
                    ranks = genome_info[1].split(";")
                    if gene_id.startswith("GB") or gene_id.startswith("RS"):
                        gene_id = gene_id[3:]
                    if gene_id.startswith("GC") or (gene_id.startswith("U") and gene_id in mapdict):
                        # temporary dictionary of IDs of interest in
                        # new_taxonomy
                        tempdict[gene_id] = {"d": ranks[0], "p": ranks[1], "c": ranks[2], "o": ranks[3],
                                             "f": ranks[4], "g": ranks[5], "s": ranks[6], 'full': genome_info[1]}
                        if tempdict.get(gene_id).get("d") != dom:
                            tempdict.pop(gene_id)

            with open(old_taxonomy) as taxold:
                for line in taxold:
                    oldnumber += 1
                    genome_info = line.strip().split("\t")
                    gene_id = genome_info[0]
                    ranks = genome_info[1].split(";")
                    if gene_id.startswith("GB") or gene_id.startswith("RS"):
                        gene_id = gene_id[3:]
                    if gene_id.startswith("GC"):
                        if gene_id in tempdict:
                            intersection += 1
                            newdict[gene_id] = tempdict.get(gene_id)
                            olddict[gene_id] = {"d": ranks[0], "p": ranks[1], "c": ranks[2], "o": ranks[3],
                                                "f": ranks[4], "g": ranks[5], "s": ranks[6], 'full': genome_info[1]}
                        else:
                            if gene_id in mapdict:
                                other_id = mapdict.get(gene_id)
                                if other_id in tempdict:
                                    intersection += 1
                                    newdict[gene_id] = tempdict.get(other_id)
                                    olddict[gene_id] = {"d": ranks[0], "p": ranks[1], "c": ranks[2], "o": ranks[3],
                                                        "f": ranks[4], "g": ranks[5], "s": ranks[6], 'full': genome_info[1]}
                    else:
                        if gene_id in mapdict:
                            other_id = mapdict.get(gene_id)
                            if other_id in tempdict:
                                intersection += 1
                                newdict[gene_id] = tempdict.get(other_id)
                                olddict[gene_id] = {"d": ranks[0], "p": ranks[1], "c": ranks[2], "o": ranks[3],
                                                    "f": ranks[4], "g": ranks[5], "s": ranks[6], 'full': genome_info[1]}
                        else:
                            if gene_id in tempdict:
                                intersection += 1
                                newdict[gene_id] = tempdict.get(gene_id)
                                olddict[gene_id] = {"d": ranks[0], "p": ranks[1], "c": ranks[2], "o": ranks[3],
                                                    "f": ranks[4], "g": ranks[5], "s": ranks[6], 'full': genome_info[1]}

            pref = 'bac'
            if dom == "d__Archaea":
                pref = "arc"

            print("Analysis of {0}").format(dom)
            print("Number of genomes in old taxonomy: {0}").format(oldnumber)
            print("Number of genomes in new taxonomy: {0}").format(newnumber)
            print("Number of genomes present in both taxonomy: {0}").format(
                intersection)
            print("Number of genomes in new taxonomy, unavailable in old taxonomy: {0}").format(
                newnumber - intersection)
            print("Number of genomes in old taxonomy, unavailable in new taxonomy: {0}").format(
                oldnumber - intersection)

            self.genomebygenomecompare(olddict, newdict, gtogfile)

            print("- Analysis of phyla")
            self.rundistrib(olddict, newdict, "p", os.path.join(output_directory, pref +
                                                                "_phylum_difference.tsv"), "phylum", keep_placeholder_name, keep_no_change, top_change)
            print("- Analysis of classes")
            self.rundistrib(olddict, newdict, "c", os.path.join(output_directory, pref +
                                                                "_class_difference.tsv"), "class", keep_placeholder_name, keep_no_change, top_change)
            print("- Analysis of orders")
            self.rundistrib(olddict, newdict, "o", os.path.join(output_directory, pref +
                                                                "_order_difference.tsv"), "order", keep_placeholder_name, keep_no_change, top_change)
            print("- Analysis of families")
            self.rundistrib(olddict, newdict, "f", os.path.join(output_directory, pref +
                                                                "_family_difference.tsv"), "family", keep_placeholder_name, keep_no_change, top_change)
            print("- Analysis of genera")
            self.rundistrib(olddict, newdict, "g", os.path.join(output_directory, pref +
                                                                "_genus_difference.tsv"), "genus", keep_placeholder_name, keep_no_change, top_change)

        gtogfile.close()

    def genomebygenomecompare(self, olddict, newdict, gtogfile):
        orderrank = ["domain", "phylum", "class",
                     'order', 'family', 'genus', 'species']
        for k, v in newdict.items():
            if k in olddict:
                if v.get('full') == olddict.get(k).get('full'):
                    gtogfile.write("{0}\t{1}\t{2}\tidentical\n".format(
                        k, olddict.get(k).get('full'), v.get('full')))
                else:
                    for rk in orderrank:
                        if v.get(rk[0:1]) != olddict.get(k).get(rk[0:1]):
                            gtogfile.write("{0}\t{1}\t{2}\t{3} difference\n".format(
                                k, olddict.get(k).get('full'), v.get('full'), rk))
            else:
                gtogfile.write(
                    "{0}\tnot_available\t{1}\t\n".format(k, v.get('full')))
        for k, v in olddict.items():
            if k not in newdict:
                gtogfile.write("{0}\t{1}\tnot_available\t\n".format(
                    k, olddict.get(k).get('full')))

    def rundistrib(self, olddict, newdict, rank, outfile, ref, keep_placeholder_name, keep_no_change, top_change):
        orderrank = ["d", "p", "c", 'o', 'f', 'g', 's']
        different_ranks = []
        # take the different ranks in the dictionaries and then count the appearances of each pair
        # of ranks ie. (p__Fusobacteria, p_Fusobacteriota),200
        for k, v in newdict.items():
            if k in olddict:
                different_ranks.append((olddict.get(k).get(rank), v.get(rank)))
        order_counter = Counter(different_ranks).most_common()
        results_dict = {}
        """create a dictionary with the name of the rank in olddict as key and the number of 
        genomes belonging to that rank encountered, along with different rank names on newdict
        for those genomes and how many times the different names appear."""
        for item in order_counter:
            if item[0][0] in results_dict:
                results_dict[item[0][0]]["nber_g"] = results_dict[item[0]
                                                                  [0]]["nber_g"] + item[1]
                results_dict[item[0][0]].get(
                    "genomes").append((item[0][1], item[1]))
            else:
                results_dict[item[0][0]] = {
                    "nber_g": item[1], "genomes": [(item[0][1], item[1])]}

        to_pop = []  # store taxonomies to take out of the file
        for k, v in results_dict.items():
            if top_change == True:
                top_gen = results_dict[k]['genomes'][0][0]
                if top_gen == k:
                    to_pop.append(k)
                else:
                    results_dict[k]['genomes'] = [
                        results_dict[k]['genomes'][0]]
            if keep_no_change == True and k == results_dict[k]['genomes'][0][0] and results_dict[k]['nber_g'] == results_dict[k]['genomes'][0][1]:
                to_pop.append(k)
            if keep_placeholder_name == True and self.hasNumber(k) == True and results_dict[k]['genomes'][0][0] == rank + "__":
                to_pop.append(k)
            elif keep_placeholder_name == True and self.hasNumber(results_dict[k]['genomes'][0][0]) == True and k == rank + "__":
                to_pop.append(k)
        for k in to_pop:
            results_dict.pop(k)

        outf = open(outfile, "w")
        outf.write(
            "Previous {0}\tNumber of genomes\tNumber of New {0}\tList of New {0}\n".format(ref))
        for k, v in results_dict.items():
            for sk, sv in olddict.items():
                if sv.get(rank) == k:
                    prankold = sv.get(orderrank[orderrank.index(rank) - 1])
                    break
            number_sub = len(v.get("genomes"))

            results = []

            if len(v.get("genomes")) == 1 and v.get("genomes")[0][0] == k:
                continue

            for newg in v.get("genomes"):
                newg_name = newg[0]
                newg_numb = float(newg[1])
                for sk, sv in newdict.items():
                    if sv.get(rank) == newg_name:
                        pranknew = sv.get(orderrank[orderrank.index(rank) - 1])
                        break
                res = newg_name
                if prankold != pranknew:
                    res = res + "(" + pranknew + ")"
                res = "{0} {1}%".format(res, round(
                    (newg_numb / v.get("nber_g")) * 100, 2))
                results.append(res)

            outf.write("{0}\t{1}\t{2}\t{3}\n".format(
                k, v.get("nber_g"), number_sub, ", ".join(results)))
        outf.close()

    def hasNumber(self, inputString):
        """check whether a taxonomy name has a number
        (likely a placeholder name)"""
        return any(char.isdigit() for char in inputString)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import os
import sys
import logging
import re

from collections import defaultdict
from collections import Counter

from biolib.newick import parse_label
from biolib.taxonomy import Taxonomy
import dendropy

class Visualisation(object):
    
    def __init__(self):
        """"Initialize"""
        self.logger = logging.getLogger()
        
    def check_change(self, old_taxonomy, new_taxonomy, output_directory):
        
        for dom in ["d__Archaea","d__Bacteria"]:
            tempdict = {} #temporary dictionary for the information in old taxonomy
            olddict = {}
            newdict = {}
            mapdict = {}# dictionary of uba_mappings
            completeold ={}
            oldnumber = 0 #number of genes in old_tax
            newnumber=0 #number of genes in new_tax
            intersection=0 #number of genes common to both taxonomies
            
            with open("uba_mapping/uba_mapping_v2.tsv", "r") as tax_map:
                for line in tax_map:
                    genes_id =line.strip().split("\t")
                    #IDs that start with U_
                    u_id = genes_id[0]
                    #IDs that start with GCA or GCF
                    gc_id = genes_id[2]
                    #delete GB or RS prefix to ensure id-consistency
                    if gc_id.startswith("GB") or gc_id.startswith("RS"):
                        gc_id = gc_id[3:]
                    #map k, v of all id_mappings
                    mapdict[gc_id]= u_id
                    mapdict[u_id]= gc_id
              
            with open(new_taxonomy) as taxnew:
                for line in taxnew:
                    newnumber+=1
                    genome_info = line.strip().split("\t")
                    gene_id = genome_info[0]
                    ranks = genome_info[1].split(";")
                    if gene_id.startswith("GB") or gene_id.startswith("RS"):
                        gene_id = gene_id[3:]
                    if gene_id.startswith("GC") or (gene_id.startswith("U") and gene_id in mapdict):
                        #temporary dictionary of IDs of interest in new_taxonomy
                        tempdict[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}
                        if tempdict.get(gene_id).get("d") != dom:
                            tempdict.pop(gene_id)
                        
            with open(old_taxonomy) as taxold:
                for line in taxold:
                    oldnumber+=1
                    genome_info = line.strip().split("\t")
                    gene_id = genome_info[0]
                    ranks = genome_info[1].split(";")
                    if gene_id.startswith("GB") or gene_id.startswith("RS"):
                        gene_id = gene_id[3:]
                    completeold[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}
                    #go over genes starting with GCA of GCF in the old tax
                    if gene_id.startswith("GC"):
                        #compare with GCA or GCF in new tax
                        if gene_id in tempdict:
                            intersection+=1
                            newdict[gene_id] = tempdict.get(gene_id)
                            olddict[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}
                        #compare with U_ id in new tax
                        else:
                            if gene_id in mapdict:
                                other_id = mapdict.get(gene_id)
                                if other_id in tempdict:
                                    intersection+=1
                                    newdict[gene_id] = tempdict.get(other_id)
                                    olddict[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}
                    else:#go over genes starting with U_ in old tax
                        #compare with GCF or GCA ids in new tax
                        if gene_id in mapdict:
                            other_id = mapdict.get(gene_id)
                            if other_id in tempdict: 
                                intersection+=1
                                newdict[gene_id] = tempdict.get(other_id)
                                olddict[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}
                        else:#compare with U_ ids in new tax
                            if gene_id in tempdict:
                                intersection+=1
                                newdict[gene_id] = tempdict.get(gene_id)
                                olddict[gene_id]={"d":ranks[0],"p":ranks[1],"c":ranks[2],"o":ranks[3],"f":ranks[4],"g":ranks[5],"s":ranks[6],'full':genome_info[1]}

                      
            pref = 'bac'
            if dom == "d__Archaea":
                pref = "arc"
				
            print("Analysis of {0}").format(dom)
            print("Number of genomes in old taxonomy: {0}").format(oldnumber)
            print("Number of genomes in new taxonomy: {0}").format(newnumber)
            print("Number of genomes present in both taxonomy: {0}").format(intersection)
            print("Number of genomes in new taxonomy, unavailable in old taxonomy: {0}").format(newnumber-intersection)
            print("Number of genomes in old taxonomy, unavailable in new taxonomy: {0}").format(oldnumber-intersection)
            
            
            print("- Analysis of phyla")
            self.run_change(olddict,newdict,"p",os.path.join(output_directory,pref+"_phylum_difference_evaluation.tsv"),"phylum",completeold)
            print("- Analysis of classes")
            self.run_change(olddict,newdict,"c",os.path.join(output_directory,pref+"_class_difference_evaluation.tsv"),"class",completeold)
            print("- Analysis of orders")
            self.run_change(olddict,newdict,"o",os.path.join(output_directory,pref+"_order_difference_evualuation.tsv"),"order",completeold)
            print("- Analysis of families")
            self.run_change(olddict,newdict,"f",os.path.join(output_directory,pref+"_family_difference_evaluation.tsv"),"family",completeold)
            print("- Analysis of genera")
            self.run_change(olddict,newdict,"g",os.path.join(output_directory,pref+"_genus_difference_evaluation.tsv"),"genus",completeold)

    def run_change(self, olddict, newdict, rank, outfile, ref, complete_old):
        
        different_ranks = []
        conserved_ranks=[]
        tax1_ranks=[]
        tax2_ranks=[]
        full_old={}
        #take the different ranks in the dictionaries and then count the appearances of each pair
        #of ranks ie. ((p__Fusobacteria, p_Fusobacteriota),200)
        for k,v in newdict.items():
            tax2_ranks.append(v.get(rank))
            if k in olddict:
                tax1_ranks.append(olddict.get(k).get(rank))
                if olddict.get(k).get(rank) != v.get(rank) and olddict.get(k).get(rank)!=rank+'__' and v.get(rank)!=rank+'__': 
                    different_ranks.append((olddict.get(k).get(rank),v.get(rank)))
                    if rank=='c':
                        full_old[olddict.get(k).get(rank)]=(olddict.get(k).get('p'))
                    elif rank =='o':
                        full_old[olddict.get(k).get(rank)]=[olddict.get(k).get('p'), olddict.get(k).get('c')]
                    elif rank=='f':
                        full_old[olddict.get(k).get(rank)]=[olddict.get(k).get('p'), olddict.get(k).get('c'),olddict.get(k).get('o')]
                    elif rank=='g':
                        full_old[olddict.get(k).get(rank)]=[olddict.get(k).get('p'), olddict.get(k).get('c'),olddict.get(k).get('o'),olddict.get(k).get('f')]
                elif olddict.get(k).get(rank) == v.get(rank):
                    conserved_ranks.append((olddict.get(k).get(rank),v.get(rank)))
        
        tax1_counter = Counter(tax1_ranks).most_common()
        tax2_counter = Counter(tax2_ranks).most_common()
        order_counter =  Counter(different_ranks).most_common()
        conserved_counter = Counter(conserved_ranks).most_common()
        #print(order_counter)
        """assign change evaluations to all pairs of rank name changes"""
        assigndict={}
        for v in complete_old.values():
            for item in order_counter:
                #check change placeholder to placeholder, invalid name, valid name
                if self.isPlaceholder(item[0][0]) and item[0]not in assigndict:
                    if self.isPlaceholder(item[0][1]):
                        assigndict[item[0]]="Modified placeholder name"
                    elif item[0][1]==v.get(rank):
                        assigndict[item[0]]="Changed placeholder to taxa seen before"
                #check change name to placeholder, invalid name, valid name
                elif (not self.isPlaceholder(item[0][0])) and item[0]not in assigndict:
                    if self.isPlaceholder(item[0][1]):
                        assigndict[item[0]]="Changed to placeholder"
                    elif item[0][1]==v.get(rank):
                        assigndict[item[0]]="BAD"
        for item in order_counter:
            if item[0] not in assigndict:
                assigndict[item[0]]="OK"
 
            
        outf= open(outfile, 'w')
        outf.write("Previous Taxonomy\tPrevious {0}\tNew {0}\tChange Evaluation\tChanged total\tTotal in previous\tTotal in new\tConserved Total\n".format(ref))
        for item in order_counter:
            prev=item[0][0]
            new=item[0][1]
            c_eval=assigndict.get(item[0])
            changed_g=item[1]
            nber_conserved=0
            prev_tax=''
            if full_old.get(prev):
                for item in full_old.get(prev):
                    prev_tax = prev_tax+item+";"
            prev_tax= prev_tax[0:-2]
            #prev_tax = full_old.get(prev)
            for item in tax1_counter:
                if item[0]==prev:
                    nber_tax1=item[1]
            for item in tax2_counter:
                if item[0]==new:
                    nber_tax2=item[1]
            for item in conserved_counter:
                if item[0][0]==prev:
                    nber_conserved=item[1]
            
            outf.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\n".format(prev_tax,prev,new,c_eval,changed_g,nber_tax1,nber_tax2,nber_conserved))
        outf.close()
    def isPlaceholder(self, inputString):
        """check whether a taxonomy name has a number
        (likely a placeholder name)"""
        return any(char.isdigit() for char in inputString)



#!/usr/bin/env python

###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import os
import sys
import re
import operator
import logging
from collections import defaultdict, Counter

from biolib.taxonomy import Taxonomy

from gtdb_validation_tk.common import read_ncbi_taxonomy


class ValidateSpeciesClusters(object):
    """Verify classification of genomes assigned to the same species cluster."""

    def __init__(self):
        """Initialization."""
        
        self.logger = logging.getLogger()
        
    def _species_clusters(self, gtdb_metadata_file):
        """Get GTDB species clusters."""
        
        self.logger.info('Identifying species clusters.')
        sp_clusters = defaultdict(set)
        genome_rep = {}
        gtdb_reps = set()
        with open(gtdb_metadata_file) as f:
            header = f.readline().strip().split('\t')

            gtdb_genome_rep_index = header.index('gtdb_representative')
            gtdb_gid_rep_index = header.index('gtdb_genome_representative')
            
            for line in f:
                line_split = line.strip().split('\t')
                
                gid = line_split[0]
                gtdb_rep = line_split[gtdb_genome_rep_index]
                gtdb_rep_gid = line_split[gtdb_gid_rep_index]
                
                if gtdb_rep == 't':
                    gtdb_reps.add(gid)
                
                if gtdb_rep_gid and gtdb_rep_gid != 'none':
                    genome_rep[gid] = gtdb_rep_gid
                    sp_clusters[gtdb_rep_gid].add(gid)
                    
        self.logger.info('Identified %d species clusters spanning %d genomes.' % (
                            len(gtdb_reps),
                            len(genome_rep)))
        
        assert(len(sp_clusters) == len(gtdb_reps))
                            
        return sp_clusters
        
    def _check_sp_assignments(self, sp_clusters, gtdb_taxonomy, output_dir):
        """Check that all genomes in a cluster have the same name as their representative."""
        
        fout = open(os.path.join(output_dir, 'check_sp_clusters.tsv'), 'w')
        fout.write('Representative ID\tGenome ID\tRepresentative taxonomy\tGenome taxonomy\n')
        num_failed = 0
        for rid in sp_clusters:
            for gid in sp_clusters[rid]:
                if gtdb_taxonomy[rid] != gtdb_taxonomy[gid]:
                    fout.write('%s\t%s\t%s\t%s\n' % (rid, gid, gtdb_taxonomy[rid], gtdb_taxonomy[gid]))
                    num_failed += 1
                    
        fout.close()
                
        if num_failed != 0:
            self.logger.info('Identified %d cases of genomes with conflicting species assignments.' % num_failed)
        else:
            self.logger.info('All genomes in taxonomy file passed validation.')

    def _check_sp_genus_name(self, gtdb_taxonomy, output_dir):
        """Check genus name of species is same as genus."""
        
        fout = open(os.path.join(output_dir, 'check_sp_genus_name.tsv'), 'w')
        fout.write('Genome ID\tGenus\tSpecies\n')
        num_failed = 0
        for gid, taxa in gtdb_taxonomy.items():
            if taxa[6] == 's__':
                continue # skip since prior to R86 this was not uncommon
                
            genus = taxa[5][3:]
            sp = taxa[6][3:]
            
            if not sp.startswith(genus):
                fout.write('%s\t%s\t%s\n' % (gid, taxa[5], taxa[6]))
                num_failed += 1

        fout.close()
        
        if num_failed != 0:
            self.logger.info('Identified %d cases of genomes with conflicting genus names.' % num_failed)
        else:
            self.logger.info('All genomes in taxonomy file passed genus name check.')
            
    def _check_gtdb_genera(self, sp_clusters, gtdb_taxonomy, ncbi_taxonomy, output_dir):
        """Check if GTDB genera have support from NCBI taxonomy."""
        
        # get all genomes in GTDB genus
        genus_gids = defaultdict(set)
        for gid, taxa in gtdb_taxonomy.items():
            genus_gids[taxa[5]].add(gid)
            
            if gid in sp_clusters:
                for cid in sp_clusters[gid]:
                    genus_gids[taxa[5]].add(cid)
                    
        # check if NCBI taxonomy supports genus assignment
        fout = open(os.path.join(output_dir, 'check_genera.tsv'), 'w')
        fout.write('GTDB genus\tMost common NCBI genus\tMost common percentage\tNCBI genera\tGenus size\tGenome IDs\n')
            
        num_failed = 0
        for gtdb_genus, gids in genus_gids.items():
            # check if specific name is a placeholder name
            is_placeholder = False
            for idx, ch in enumerate(gtdb_genus[3:]):
                if not ch.isalpha() or (idx>0 and not ch.islower()):
                    # genus contains a non-alphabetic and/or 
                    # an uppercase character besides the initial
                    # capitalization so is a placeholder name
                    is_placeholder = True
                    break

            if is_placeholder:
                continue
                
            # create canonical genus name
            canonical_gtdb_genus = gtdb_genus
            if '_' in canonical_gtdb_genus[3:]:
                canonical_gtdb_genus = canonical_gtdb_genus[0:canonical_gtdb_genus.rfind('_')]
        
            # check for support in NCBI taxonomy
            ncbi_genera = []
            for gid in gids:
                ncbi_genus = ncbi_taxonomy[gid][5]
                if ncbi_genus == 'g__':
                    continue 
                    
                ncbi_genus = ncbi_genus.replace('Candidatus ', '')
                ncbi_genera.append(ncbi_genus)
                
            if ncbi_genera:
                # consider all genera that are equally abundant
                ncbi_counter = Counter(ncbi_genera)
                prev_genus_count = 0
                for ncbi_genus_common, genus_count in ncbi_counter.most_common():
                    if genus_count < prev_genus_count:
                        break # no longer at equally abundant species name
                        
                    prev_genus_count = genus_count
                        
                    if ncbi_genus_common != canonical_gtdb_genus:
                        num_failed += 1
                        fout.write('%s\t%s\t%.2f\t%s\t%d\t%s\n' % (
                            gtdb_genus,
                            ncbi_genus_common,
                            float(genus_count)*100.0/len(ncbi_genera),
                            ', '.join(['%s:%d' % (k,v) for k,v in ncbi_counter.most_common()]),
                            len(gids),
                            ','.join(gids)))
            else:
                # no genomes with an assigned NCBI genus
                num_failed += 1
                fout.write('%s\t%s\t%.2f\t%s\t%d\t%s\n' % (
                            gtdb_genus,
                            'no assigned genera',
                            0.0,
                            '',
                            len(gids),
                            ','.join(gids)))

        fout.close()
                    
        if num_failed != 0:
            self.logger.info('Identified %d cases of potentially unsupported GTDB genus assignments.' % num_failed)
        else:
            self.logger.info('All GTDB genus assignments appear to have well-supported by the NCBI taxonomy.')
            
    def _check_sp_specific_name(self, sp_clusters, gtdb_taxonomy, ncbi_taxonomy, output_dir):
        """Check is specific name of GTDB species has support from NCBI species assignments."""

        fout = open(os.path.join(output_dir, 'check_sp_specific_name.tsv'), 'w')
        fout.write('Representative ID\tCluster size\tGTDB species')
        fout.write('\tMost common NCBI species\tMost common percentage\tNCBI species\tCluster IDs\n')
            
        num_failed = 0
        for rid, cluster_ids in sp_clusters.items():
            gtdb_sp = gtdb_taxonomy[rid][6]
            if gtdb_sp == 's__':
                continue # skip since prior to R86 this was not uncommon
                
            # get canonical version of specific name
            gtdb_specific_name = gtdb_sp.split()[-1]
            if '_' in gtdb_specific_name:
                gtdb_specific_name = gtdb_specific_name[0:gtdb_specific_name.rfind('_')]
            
            # check if specific name is a placeholder name
            is_placeholder = False
            for ch in gtdb_specific_name:
                if not ch.isalpha() or not ch.islower():
                    # specific name contains a non-alphabetic and/or 
                    # uppercase character so is a placeholder name
                    is_placeholder = True
                    break

            if is_placeholder:
                continue

            # check if NCBI taxonomy agrees with GTDB specific name
            ncbi_sp_names = []
            for gid in cluster_ids.union([rid]):
                ncbi_sp = ncbi_taxonomy[gid][6]
                if ncbi_sp == 's__':
                    continue 
                    
                ncbi_sp_names.append(ncbi_sp)
                
            if ncbi_sp_names:
                # consider all species names that are equally abundant
                ncbi_counter = Counter(ncbi_sp_names)
                prev_sp_count = 0
                for ncbi_sp_common, sp_count in ncbi_counter.most_common():
                    if sp_count < prev_sp_count:
                        break # no longer at equally abundant species name
                        
                    prev_sp_count = sp_count
                        
                    ncbi_specific_name = ncbi_sp_common.split()[-1]
                    if ncbi_specific_name != gtdb_specific_name:
                        num_failed += 1
                        fout.write('%s\t%d\t%s\t%s\t%.2f\t%s\t%s\n' % (
                                rid,
                                len(cluster_ids),
                                gtdb_sp,
                                ncbi_sp_common,
                                float(sp_count)*100.0/len(ncbi_sp_names),
                                ', '.join(['%s:%d' % (k,v) for k,v in ncbi_counter.most_common()]),
                                ','.join(cluster_ids)))
            else:
                # no genomes with an assigned NCBI species name
                num_failed += 1
                fout.write('%s\t%d\t%s\t%s\t%.2f\t%s\t%s\n' % (
                            rid,
                            len(cluster_ids),
                            gtdb_sp,
                            'no assigned species',
                            0.0,
                            '',
                            ','.join(cluster_ids)))

        fout.close()
        
        if num_failed != 0:
            self.logger.info('Identified %d cases of potentially unsupported specific species names.' % num_failed)
        else:
            self.logger.info('All clusters appear to have well-supported specific species names.')
            
    def run(self, input_taxonomy, gtdb_metadata_file, output_dir):
        """Verify classification of genomes assigned to the same species cluster."""
        
        gtdb_taxonomy = Taxonomy().read(input_taxonomy)
        ncbi_taxonomy = read_ncbi_taxonomy(gtdb_metadata_file)
        
        # check that all genomes have an assigned species 
        for gid, taxa in gtdb_taxonomy.items():
            if taxa[6] == 's__':
                # issue warning since this was common prior to R89
                self.logger.warning('Genome %s does not have an assigned species.' % gid)
                
            if taxa[5] == 'g__':
                self.logger.error('Genome %s does not have an assigned genus.' % gid)
                sys.exit(-1)
        
        # identify clusters
        sp_clusters = self._species_clusters(gtdb_metadata_file)

        # validate that genomes in the same species cluster have the same assignment
        self._check_sp_assignments(sp_clusters, gtdb_taxonomy, output_dir)
        
        # check if GTDB genera have support from NCBI taxonomy
        self._check_gtdb_genera(sp_clusters, gtdb_taxonomy, ncbi_taxonomy, output_dir)
        
        # validate genus name of species assignments
        self._check_sp_genus_name(gtdb_taxonomy, output_dir)
        
        # check is specific name of GTDB species has support from NCBI species assignments
        self._check_sp_specific_name(sp_clusters, gtdb_taxonomy, ncbi_taxonomy, output_dir)




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #r
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Donovan Parks"
__copyright__ = "Copyright 2017"
__credits__ = ["Donovan Parks"]
__license__ = "GPL3"
__maintainer__ = "Donovan Parks"
__email__ = "donovan.parks@gmail.com"
__status__ = "Development"

import argparse
import logging
import ntpath
import os
import sys
from biolib.common import make_sure_path_exists
from biolib.misc.custom_help_formatter import CustomHelpFormatter

from gtdb_validation_tk import __version__
from gtdb_validation_tk.main import OptionsParser


def print_help():
    """Help menu."""

    print('')
    print('                ...::: GTDB Validation Toolkit v' + __version__ + ' :::...''')
    print('''\

    Taxonomy validation:
      check_file      -> Verify formatting of taxonomy file and check taxon names
      check_tree      -> Verify taxonomy of decorated tree and check for polyphyletic groups
      red_check       -> Report taxa with unexpected RED values
      missing_genera  -> Report lineages that may represent missing genus names
      merge_genera    -> Report polyphyletic genera that are candidates for merging
      support         -> Report taxa with low support values 
      check_spelling  -> Report taxa with potential typos based on previous taxonomy

    Tree validation
      branch_len -> Calculate the mean branch length from root to extant taxa (expected to be ~1.0)

    Species validation:
      check_generic    -> Verify that generic name of species is valid
      check_specific   -> Verify that specific name of species is valid
      species_clusters -> Verify classification of genomes assigned to the same species cluster
      type_strains     -> Compare GTDB to type strain of species designations at NCBI, LPSN, DSMZ, and/or StrainInfo
      type_species     -> Compare GTDB to type species of genus designations at NCBI, LPSN, DSMZ, and/or StrainInfo

    Species comparison:
      ncbi_species      -> Report differences between GTDB and NCBI species names
      gtdb_cluster_ncbi -> Report NCBI species within each GTDB-defined cluster

    RED comparison:
      red_diff  -> Report RED values accross 2 GTDB releases

    Taxonomy comparison:
      check_distrib -> Report changes of genomes distribution between different taxonomies
      tax_diff      -> Determine differences between two taxonomy files
      tree_diff     -> Report bootstrap supported topological differences between trees
      retired_taxa  -> Report all taxa that have been retired (lost?) between two taxonomy files

    Curation tree methods:
      add_dummy     -> Add a zero branch length sister node to each current leaf node
      remove_dummy  -> Remove dummy nodes from curation tree

    General curation methods:
      val_table     -> Validation table containing useful information for all representative genomes
      suffix_table  -> Create table indicating last polyphyly suffix used for each taxon
      ncbi_tax_diff -> Establish changes in NCBI taxonomy that may impact GTDB taxonomy

  Use: gtdb_validation_tk <command> -h for command specific help.

  Feature requests or bug reports can be sent to Donovan Parks (donovan.parks@gmail.com)
    or posted on GitHub (https://github.com/Ecogenomics/gtdb_validation_tk).
    ''')


def logger_setup(output_dir, silent):
    """Set logging for application.

    Parameters
    ----------
    output_dir : str
        Output directory for log file.
    silent : boolean
        Flag indicating if output to stdout should be suppressed.
    """

    # setup general properties of logger
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    log_format = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                   datefmt="%Y-%m-%d %H:%M:%S")

    # setup logging to console
    if not silent:
        stream_logger = logging.StreamHandler(sys.stdout)
        stream_logger.setFormatter(log_format)
        stream_logger.setLevel(logging.DEBUG)
        logger.addHandler(stream_logger)

    if output_dir:
        make_sure_path_exists(output_dir)
        file_logger = logging.FileHandler(os.path.join(output_dir, 'gtdb_validation_tk.log'), 'a')
        file_logger.setFormatter(log_format)
        logger.addHandler(file_logger)

    logger.info('GTDB Validation Toolkit v%s' % __version__)
    logger.info(ntpath.basename(sys.argv[0]) + ' ' + ' '.join(sys.argv[1:]))


def main():
    # initialize the options parser
    parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(help="--", dest='subparser_name')

    # validate taxonomy file
    check_file_parser = subparsers.add_parser('check_file',
                                              formatter_class=CustomHelpFormatter,
                                              description='Verify taxonomy file is formatted as expected.')
    check_file_parser.add_argument('input_taxonomy', help='input taxonomy file')
    check_file_parser.add_argument('--include_species', help="include species in validation checks",
                                   action='store_true')
    check_file_parser.add_argument('--silent', help="suppress output", action='store_true')

    # validate decorated tree
    check_tree_parser = subparsers.add_parser('check_tree',
                                              formatter_class=CustomHelpFormatter,
                                              description='Verify taxonomy of decorated tree and check for polyphyletic groups.')
    check_tree_parser.add_argument('decorated_tree', help='decorated tree to validate')
    check_tree_parser.add_argument('--taxonomy_file',
                                   help='alternative taxonomy to validate with respect to the input tree')
    check_tree_parser.add_argument('--include_species', help="include species in validation checks",
                                   action='store_true')
    check_tree_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check polyphyly suffix
    check_suffix_parser = subparsers.add_parser('check_suffix',
                                                formatter_class=CustomHelpFormatter,
                                                description='Report taxa with potentially erroneous polyphyly suffix designations.')
    check_suffix_parser.add_argument('input_taxonomy', help='input taxonomy file')
    check_suffix_parser.add_argument('--include_species', help="include species in validation checks",
                                     action='store_true')
    check_suffix_parser.add_argument('--silent', help="suppress output", action='store_true')

    # report taxa with unexpected RED values
    red_check_parser = subparsers.add_parser('red_check',
                                             formatter_class=CustomHelpFormatter,
                                             description='Report taxa with unexpected RED values.')
    red_check_parser.add_argument('scaled_tree',
                                  help='scaled and decorated validation tree to check for missing genus names')
    red_check_parser.add_argument('red_dictionary',
                                  help='file with RED values for validation tree (*.dict file from phylorank outlier)')
    red_check_parser.add_argument('output_file', help='output file')
    red_check_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check for potentially missing genera
    missing_genera_parser = subparsers.add_parser('missing_genera',
                                                  formatter_class=CustomHelpFormatter,
                                                  description='Report lineages that may represent missing genus names.')
    missing_genera_parser.add_argument('scaled_tree', help='scaled and decorated tree to check for missing genus names')
    missing_genera_parser.add_argument('red_dictionary',
                                       help='file with RED values (*.dict file from phylorank outlier)')
    missing_genera_parser.add_argument('gtdb_prev_taxonomy', help='GTDB taxonomy file for previous GTDB release')
    missing_genera_parser.add_argument('ncbi_taxonomy', help='NCBI taxonomy file for current GTDB release')
    missing_genera_parser.add_argument('output_file', help='output file')
    missing_genera_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check for sister genera that could be merged
    merge_genera_parser = subparsers.add_parser('merge_genera',
                                                formatter_class=CustomHelpFormatter,
                                                description='Report polyphyletic genera that are candidates for merging.')
    merge_genera_parser.add_argument('scaled_tree', help='scaled and decorated tree to check for missing genus names')
    merge_genera_parser.add_argument('red_dictionary', help='file with RED values (*.dict file from phylorank outlier)')
    merge_genera_parser.add_argument('output_file', help='output file')
    merge_genera_parser.add_argument('--silent', help="suppress output", action='store_true')

    # report taxa with low support values
    support_parser = subparsers.add_parser('support',
                                           formatter_class=CustomHelpFormatter,
                                           description='Report taxa with low support values.')
    support_parser.add_argument('decorated_tree', help='decorated tree to validate')
    support_parser.add_argument('--min_support', help='report all taxa below this support value', type=int, default=70)
    support_parser.add_argument('--silent', help="suppress output", action='store_true')

    # report RED values for 2 GTDB releases
    red_diff_parser = subparsers.add_parser('red_diff',
                                            formatter_class=CustomHelpFormatter,
                                            description='Report RED values for 2+ GTDB releases.')
    red_diff_parser.add_argument('-r', '--release', nargs=3, action='append',
                                 help='--release <version> <RED values> <dictionary>')
    red_diff_parser.add_argument('-o', '--output_file', help='Output file.')
    red_diff_parser.add_argument('--silent', help="suppress output", action='store_true')

    # difference between two taxonomy files
    tax_diff_parser = subparsers.add_parser('tax_diff',
                                            formatter_class=CustomHelpFormatter,
                                            description='Determine differences between two taxonomy files.')
    tax_diff_parser.add_argument('input_taxonomy1', help='first taxonomy file')
    tax_diff_parser.add_argument('input_taxonomy2', help='second taxonomy file')
    tax_diff_parser.add_argument('output_dir', help="output directory")
    tax_diff_parser.add_argument('--gtdb_metadata_file',
                                 help='file specifying GTDB metadata (TSV) if type material fields are desired')
    tax_diff_parser.add_argument('--silent', help="suppress output", action='store_true')

    # quanitify differences between two trees
    tree_diff_parser = subparsers.add_parser('tree_diff',
                                             formatter_class=CustomHelpFormatter,
                                             description='Quantify bootstrap supported topological differences between trees.')

    tree_diff_parser.add_argument('input_tree1', help="first decorated input tree")
    tree_diff_parser.add_argument('input_tree2', help="second decorated input tree")
    tree_diff_parser.add_argument('output_dir', help="output directory")
    tree_diff_parser.add_argument('--min_support', help="minimum value to consider a lineage well supported",
                                  type=float, default=90.0)
    tree_diff_parser.add_argument('--min_taxa', help="only consider lineage with sufficient number of taxa", type=int,
                                  default=2)
    tree_diff_parser.add_argument('--silent', help="suppress output", action='store_true')

    # difference between two taxonomy files
    retired_taxa_parser = subparsers.add_parser('retired_taxa',
                                                formatter_class=CustomHelpFormatter,
                                                description='Report all taxa that have been retired (lost?) between two taxonomy files.')
    retired_taxa_parser.add_argument('new_taxonomy_file', help='file with new taxonomy')
    retired_taxa_parser.add_argument('prev_taxonomy_file', help='file with previous taxonomy')
    retired_taxa_parser.add_argument('output_table', help="output table")
    retired_taxa_parser.add_argument('--silent', help="suppress output", action='store_true')

    # calculate mean root to tip branch length
    branch_len_parser = subparsers.add_parser('branch_len',
                                              formatter_class=CustomHelpFormatter,
                                              description='Calculate the mean branch length from root to extant taxa.')

    branch_len_parser.add_argument('input_tree', help="tree to interest")
    branch_len_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check taxonomy for spelling mistakes
    check_spelling_parser = subparsers.add_parser('check_spelling',
                                                  formatter_class=CustomHelpFormatter,
                                                  description='Report taxa with potential typos based on previous taxonomy.')
    check_spelling_parser.add_argument('input_taxonomy1', help='current taxonomy file')
    check_spelling_parser.add_argument('input_taxonomy2', help='previous, or otherwise trusted, taxonomy file')
    check_spelling_parser.add_argument('output_file', help="output directory")
    check_spelling_parser.add_argument('--dist', help='Levenshtein distance threshold', default=95)
    check_spelling_parser.add_argument('--silent', help="suppress output", action='store_true')

    # list which group has been split/modified between gtdb iterations
    check_dist_parser = subparsers.add_parser('check_distrib',
                                              formatter_class=CustomHelpFormatter,
                                              description='Check genome distribution changes between 2 taxonomies.')
    check_dist_parser.add_argument('input_taxonomy1', help='current taxonomy file')
    check_dist_parser.add_argument('input_taxonomy2', help='previous taxonomy file')
    check_dist_parser.add_argument('output_dir', help="output directory")
    check_dist_parser.add_argument('--keep_no_change',
                                   help="keep all ranks regardless of whether they are the same in both taxonomies",
                                   action='store_false', default='store_true')
    check_dist_parser.add_argument('--keep_placeholder_name',
                                   help='keep ranks that have a placeholder name->no name in the taxonomies',
                                   action='store_false', default='store_true')
    check_dist_parser.add_argument('--top_change', help='keep only top change', action='store_true',
                                   default='store_false')
    check_dist_parser.add_argument('--silent', help="suppress output", action='store_true')

    # general validation table
    val_table_parser = subparsers.add_parser('val_table',
                                             formatter_class=CustomHelpFormatter,
                                             description='Validation table containing useful information for all representative genomes.')
    val_table_parser.add_argument('input_taxonomy', help='input taxonomy file')
    val_table_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    val_table_parser.add_argument('output_table', help='output table')
    val_table_parser.add_argument('--silent', help="suppress output", action='store_true')

    # suffix table
    suffix_table_parser = subparsers.add_parser('suffix_table',
                                                formatter_class=CustomHelpFormatter,
                                                description='Create table indicating last polyphyly suffix used for each taxon.')
    suffix_table_parser.add_argument('output_table', help="output directory")
    suffix_table_parser.add_argument('--input_taxonomy',
                                     help="additional taxonomy table to use for establishing suffix table")
    suffix_table_parser.add_argument('--silent', help="suppress output", action='store_true')

    # establish changes in NCBI taxonomy that may impact GTDB taxonomy
    ncbi_tax_diff_parser = subparsers.add_parser('ncbi_tax_diff',
                                                 formatter_class=CustomHelpFormatter,
                                                 description='Establish changes in NCBI taxonomy that may impact GTDB taxonomy.')
    ncbi_tax_diff_parser.add_argument('prev_ncbi_taxonomy', help='NCBI taxonomy for previous GTDB release')
    ncbi_tax_diff_parser.add_argument('cur_ncbi_taxonomy', help='NCBI taxonomy for current GTDB release')
    ncbi_tax_diff_parser.add_argument('gtdb_taxonomy', help='GTDB taxonomy for current release')
    ncbi_tax_diff_parser.add_argument('output_dir', help="output directory")
    ncbi_tax_diff_parser.add_argument('--silent', help="suppress output", action='store_true')

    # verify generic species names
    check_generic_parser = subparsers.add_parser('check_generic',
                                                 formatter_class=CustomHelpFormatter,
                                                 description='Verify that generic name of species is valid')
    check_generic_parser.add_argument('input_taxonomy', help='input taxonomy file')
    check_generic_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    check_generic_parser.add_argument('sp_classification_ledger',
                                      help='file indicating desired species names for specific genomes')
    check_generic_parser.add_argument('genus_ledger', help='file indicating desired genus names for specific genomes')
    check_generic_parser.add_argument('gtdb_synonym_file', help='file specifying GTDB declared synonyms')
    check_generic_parser.add_argument('user_genome_id_table',
                                      help='table specifying mapping between GTDB user, UBA, and GenBank IDs (TSV)')
    check_generic_parser.add_argument('output_dir', help="output directory")
    check_generic_parser.add_argument('--silent', help="suppress output", action='store_true')

    # verify specific species names
    check_specific_parser = subparsers.add_parser('check_specific',
                                                  formatter_class=CustomHelpFormatter,
                                                  description='Verify that specific name of species is valid')
    check_specific_parser.add_argument('input_taxonomy', help='input taxonomy file')
    check_specific_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    check_specific_parser.add_argument('sp_classification_ledger',
                                       help='file indicating desired species names for specific genomes')
    check_specific_parser.add_argument('gtdb_synonym_file', help='file specifying GTDB declared synonyms')
    check_specific_parser.add_argument('user_genome_id_table',
                                       help='table specifying mapping between GTDB user, UBA, and GenBank IDs (TSV)')
    check_specific_parser.add_argument('output_dir', help="output directory")
    check_specific_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check species clusters
    sp_clusters_parser = subparsers.add_parser('species_clusters',
                                               formatter_class=CustomHelpFormatter,
                                               description='Verify classification of genomes assigned to the same species cluster.')
    sp_clusters_parser.add_argument('input_taxonomy', help='input taxonomy file')
    sp_clusters_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    sp_clusters_parser.add_argument('output_dir', help="output directory")
    sp_clusters_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check type strains
    type_strains_parser = subparsers.add_parser('type_strains',
                                                formatter_class=CustomHelpFormatter,
                                                description='Compare GTDB to type strain of species designations at NCBI, LPSN, DSMZ, and/or StrainInfo.')
    type_strains_parser.add_argument('input_taxonomy', help='input taxonomy file')
    type_strains_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    type_strains_parser.add_argument('sp_classification_ledger',
                                     help='file indicating desired species names for specific genomes')
    type_strains_parser.add_argument('gtdb_named_type_genomes',
                                     help='file explicitly indicating type genomes for named species (gtdb_type_genomes_final.tsv from select_type_genomes)')
    type_strains_parser.add_argument('gtdb_synonym_file', help='file specifying GTDB declared synonyms')
    type_strains_parser.add_argument('output_dir', help='output directory')
    type_strains_parser.add_argument('--show_synonyms', help="show modified type material resulting from GTDB synonyms",
                                     action='store_true')
    type_strains_parser.add_argument('--prefix', help="prefix for all output files", default='gvtk')
    type_strains_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check type species
    type_species_parser = subparsers.add_parser('type_species',
                                                formatter_class=CustomHelpFormatter,
                                                description='Compare GTDB to type species of genus designations at NCBI, LPSN, DSMZ, and/or StrainInfo.')
    type_species_parser.add_argument('input_taxonomy', help='input taxonomy file')
    type_species_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    type_species_parser.add_argument('gtdb_synonym_file', help='file specifying GTDB declared synonyms')
    type_species_parser.add_argument('sp_classification_ledger',
                                     help='file indicating desired species names for specific genomes')
    type_species_parser.add_argument('genus_classification_ledger',
                                     help='file indicating desired genus names for specific genomes')
    type_species_parser.add_argument('output_dir', help='output directory')
    type_species_parser.add_argument('--show_synonyms', help="show modified type material resulting from GTDB synonyms",
                                     action='store_true')
    type_species_parser.add_argument('--prefix', help="prefix for all output files", default='gvtk')
    type_species_parser.add_argument('--silent', help="suppress output", action='store_true')

    # compare NCBI and GTDB species names
    ncbi_species_parser = subparsers.add_parser('ncbi_species',
                                                formatter_class=CustomHelpFormatter,
                                                description='Report differences between GTDB and NCBI species names.')
    ncbi_species_parser.add_argument('input_taxonomy', help='input taxonomy file')
    ncbi_species_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    ncbi_species_parser.add_argument('output_dir', help='output directory')
    ncbi_species_parser.add_argument('--silent', help="suppress output", action='store_true')

    # tabulate NCBI species in each GTDB-defined cluster
    gtdb_cluster_ncbi_parser = subparsers.add_parser('gtdb_cluster_ncbi',
                                                     formatter_class=CustomHelpFormatter,
                                                     description='Report NCBI species within each GTDB-defined cluster.')
    gtdb_cluster_ncbi_parser.add_argument('input_taxonomy', help='input taxonomy file')
    gtdb_cluster_ncbi_parser.add_argument('gtdb_metadata_file', help='file specifying GTDB metadata (TSV)')
    gtdb_cluster_ncbi_parser.add_argument('output_dir', help='output directory')
    gtdb_cluster_ncbi_parser.add_argument('--silent', help="suppress output", action='store_true')

    # check genome distribution changes between two GTDB releases
    check_change_parser = subparsers.add_parser('check_change',
                                                formatter_class=CustomHelpFormatter,
                                                description='Check genome distribution changes between two GTDB releases.')
    check_change_parser.add_argument('input_taxonomy1', help='current taxonomy file')
    check_change_parser.add_argument('input_taxonomy2', help='previous taxonomy file')
    check_change_parser.add_argument('output_dir', help='output directory')
    check_change_parser.add_argument('--silent', help='suppress output', action='store_true')

    # add a zero branch length sister node to each current leaf node
    add_dummy_parser = subparsers.add_parser('add_dummy',
                                             formatter_class=CustomHelpFormatter,
                                             description='Add a zero branch length sister node to each current leaf node.')
    add_dummy_parser.add_argument('input_tree', help='input tree file')
    add_dummy_parser.add_argument('--taxonomy_file', help='add dummy nodes to taxonomy file')
    add_dummy_parser.add_argument('--silent', help='suppress output', action='store_true')

    # remove dummy nodes from curation tree
    remove_dummy_parser = subparsers.add_parser('remove_dummy',
                                                formatter_class=CustomHelpFormatter,
                                                description='Remove dummy nodes from curation tree.')
    remove_dummy_parser.add_argument('input_tree', help='input tree file')
    remove_dummy_parser.add_argument('output_tree', help='output tree file')
    remove_dummy_parser.add_argument('--silent', help='suppress output', action='store_true')

    # get and check options
    args = None
    if len(sys.argv) == 1 or sys.argv[1] == '-h' or sys.argv == '--help':
        print_help()
        sys.exit(0)
    else:
        args = parser.parse_args()

    try:
        logger_setup(args.output_dir, args.silent)
    except:
        logger_setup(None, args.silent)

    # do what we came here to do
    try:
        parser = OptionsParser()
        if False:
            # import pstats
            # p = pstats.Stats('prof')
            # p.sort_stats('cumulative').print_stats(10)
            # p.sort_stats('time').print_stats(10)
            import cProfile

            cProfile.run('parser.parse_options(args)', 'prof')
        elif False:
            import pdb

            pdb.run(parser.parse_options(args))
        else:
            parser.parse_options(args)
    except SystemExit:
        print("\n  Controlled exit resulting from an unrecoverable error or warning.")
    except:
        print("\nUnexpected error:", sys.exc_info()[0])
        raise


if __name__ == '__main__':
    main()



import os
from distutils.core import setup


def version():
    setupDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(setupDir, 'gtdb_validation_tk', 'VERSION'))
    return versionFile.readline().strip()


setup(
    name='gtdb_validation_tk',
    version=version(),
    author='Donovan Parks',
    author_email='donovan.parks@gmail.com',
    packages=['gtdb_validation_tk'],
    package_data={'gtdb_validation_tk': ['VERSION']},
    entry_points={
        'console_scripts': [
            'gtdb_validation_tk = gtdb_validation_tk.__main__:main'
        ]
    },
    url='https://pypi.org/project/gtdb-validation-tk/',
    license='GPL3',
    description='A toolbox for validating the GTDB taxonomy.',
    install_requires=[
        "numpy >= 1.8.0",
        "biolib >= 0.1.6",
        "dendropy >= 4.0.0",
        'fuzzywuzzy'],
)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import re
from collections import defaultdict
from collections import Counter

from biolib.taxonomy import Taxonomy

import dendropy

def suffix_value(suffix):
    """Get value of taxa suffix.
    
    A = 65, B = 66, ..., Z = 90
    AA = 16705, AB = 16705, BA = 16961
    """
    
    v = 0
    for idx, ch in enumerate(suffix):
        v += 256**(len(suffix)-idx-1) * ord(ch)
    
    return v
    
def increment_suffix(suffix):
    """Increment suffix by one character."""
    
    if len(suffix) > 2:
        print('[Error] Unable to handle suffixes with >2 characters.')
        sys.exit(-1)
    
    if len(suffix) == 1:
        if suffix != 'Z':
            suffix = chr(ord(suffix) + 1)
        else:
            suffix = 'AA'
    else:
        last_ch = suffix[1]
        if last_ch != 'Z':
            suffix = suffix[0] + chr(ord(last_ch) + 1)
        else:
            first_ch = suffix[0]
            first_ch = chr(ord(first_ch) + 1)
            suffix = first_ch + 'A'

    return suffix


class SuffixTable(object):
    """Create table indicating last polyphyly suffix used for each taxon."""

    def __init__(self):
        """Initialize."""
        
        self.prev_taxonomy_dir = '/srv/projects/gtdb/data/taxonomy_gtdb'
        self.prev_tree_dir = '/srv/projects/gtdb/data/trees_gtdb'
        
        self.logger = logging.getLogger('timestamp')

    def run(self, input_taxonomy, output_table):
        """Create table indicating last polyphyly suffix used for each taxon."""
        
        # get all previous taxonomy files
        self.logger.info('Reading previous GTDB taxonomy files in %s:' % self.prev_taxonomy_dir)
        prev_taxonomy_files = []
        taxonomies = defaultdict(lambda: {})
        for f in os.listdir(self.prev_taxonomy_dir):
            if f.endswith('.tsv') and 'gtdb' in f:
                self.logger.info('  %s' % f)
                taxonomy_file = os.path.join(self.prev_taxonomy_dir, f)
                prev_taxonomy_files.append(taxonomy_file)
                
                taxonomy_id ='_'.join(f.split('_')[0:2])
                taxonomies[taxonomy_id].update(Taxonomy().read(taxonomy_file))
            
        self.logger.info('Considering taxonomy from %d previous releases.' % len(taxonomies))
        
        # get canonical genomes in previous GTDB reference trees
        self.logger.info('Determining canonical genomes in previous GTDB trees: %s' % self.prev_tree_dir)
        prev_canonical_gids = defaultdict(set)
        for f in os.listdir(self.prev_tree_dir):
            if f.endswith('.tree'):
                tree = dendropy.Tree.get_from_path(os.path.join(self.prev_tree_dir, f), 
                                                    schema='newick', 
                                                    rooting="force-rooted", 
                                                    preserve_underscores=True)
                gids = set()
                for leaf in tree.leaf_node_iter():
                    gids.add(leaf.taxon.label)
                    
                self.logger.info('  %s contains %d genomes' % (f, len(gids)))
                
                tree_id = '_'.join(f.split('_')[0:2])
                prev_canonical_gids[tree_id].update(gids)
                
        self.logger.info('Considering trees from %d previous GTDB releases.' % len(prev_canonical_gids))

        # get highest alphabetic suffix for each taxon
        self.logger.info('Determining highest polyphyletic alphabetic suffix for each taxon.')
        taxonomy_files = prev_taxonomy_files
        if input_taxonomy:
            self.logger.info('Also considering specified taxonomy files: %s' % input_taxonomy)
            taxonomies['input_taxonomy'] = Taxonomy().read(input_taxonomy)
            
        taxon_suffix = {}
        for taxonomy in taxonomies.values():
            for taxa in taxonomy.values():
                for taxon in taxa:
                    rank_prefix = taxon[0:3]
                    taxon_name = taxon[3:]

                    if '_' in taxon_name:
                        if rank_prefix != 's__':
                            taxon_name, suffix = taxon_name.rsplit('_', 1)
                        else:
                            # check is the specific name has a suffix
                            specific_name = taxon_name.split()[-1]
                            if '_' in specific_name:
                                taxon_name, suffix = taxon_name.rsplit('_', 1)
                            else:
                                continue

                        canonical_taxon = '%s%s' % (rank_prefix, taxon_name)
                        cur_suffix = taxon_suffix.get(canonical_taxon, 'A')
                        if suffix_value(suffix) >= suffix_value(cur_suffix):
                            taxon_suffix[canonical_taxon] = suffix
                            
        # get canonical genome for each highest suffix taxon
        self.logger.info('Determining canonical exemplar genome for each taxon.')
        example_gids = defaultdict(lambda: {})
        for tree_id in prev_canonical_gids:
            if tree_id not in taxonomies:
                continue
                
            taxonomy = taxonomies[tree_id]
            taxon_exemplars = {}
            for gid, taxa in taxonomy.items():
                for taxon in taxa:
                    taxon_exemplars[taxon] = gid
            
            for taxon, suffix in taxon_suffix.items():
                taxon_name = '%s_%s' % (taxon, taxon_suffix[taxon])
                if taxon_name in taxon_exemplars:
                    example_gids[tree_id][taxon_name] = taxon_exemplars[taxon_name]

        # write out suffix table
        if output_table:
            fout = open(output_table, 'w')
            fout.write('Taxon\tSuffix')
            for tree_id in sorted(prev_canonical_gids, reverse=True):
                fout.write('\t%s' % tree_id)
            fout.write('\n')
            
            for taxon in Taxonomy().sort_taxa(taxon_suffix):
                fout.write('%s\t%s' % (taxon, taxon_suffix[taxon]))
                
                taxon_name = '%s_%s' % (taxon, taxon_suffix[taxon])
                for tree_id in sorted(prev_canonical_gids, reverse=True):
                    if tree_id in example_gids:
                        fout.write('\t%s' % example_gids[tree_id].get(taxon_name, 'n/a'))
                    else:
                        fout.write('\tn/a')
                fout.write('\n')
                
            fout.close()
        
        return taxon_suffix


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import re
from collections import defaultdict

from biolib.taxonomy import Taxonomy
from biolib.common import canonical_gid

from gtdb_validation_tk.common import (placeholder_genus,
                                        canonical_taxon_name,
                                        latinized_specific_sp_name)


class CheckTaxonomyFile(object):
    """Validate different aspects of standard taxonomy file."""

    def __init__(self, dpi=96):
        """Initialize."""
        
        self.logger = logging.getLogger()
        
    def check(self, taxonomy,
                    check_prefixes, 
                    check_ranks, 
                    check_hierarchy, 
                    check_species, 
                    check_group_names,
                    check_duplicate_names,
                    check_capitalization):
        """Check if taxonomy forms a strict hierarchy with all expected ranks.

        Parameters
        ----------
        taxonomy : d[unique_id] -> [d__<taxon>; ...; s__<taxon>]
            Taxonomy strings indexed by unique ids.
        check_prefixes : boolean
            Flag indicating if prefix of taxon should be validated.
        check_ranks : boolean
            Flag indicating if the presence of all ranks should be validated.
        check_hierarchy : boolean
            Flag indicating if the taxonomic hierarchy should be validated.
        check_species : boolean
            Flag indicating if the taxonomic consistency of named species should be validated.
        check_group_names : boolean
            Flag indicating if group names should be checked for invalid characters.
        check_duplicate_names : boolean
            Flag indicating if group names should be checked for duplicates.
        report_errors : boolean
            Flag indicating if errors should be written to screen.

        Returns
        -------
        dict : d[taxon_id] -> taxonomy
            Taxa with invalid number of ranks.
        dict : d[taxon_id] -> [taxon, taxonomy]
            Taxa with invalid rank prefixes.
        dict: d[taxon_id] -> [species name, error message]
            Taxa with invalid species names.
        dict: d[child_taxon_id] -> two or more parent taxon ids
            Taxa with invalid hierarchies.
        """
        
        # check for incomplete taxonomy strings or unexpected rank prefixes
        Taxonomy().validate(taxonomy,
                            check_prefixes=check_prefixes,
                            check_ranks=check_ranks,
                            check_hierarchy=check_hierarchy,
                            check_species=check_species,
                            check_group_names=check_group_names,
                            check_duplicate_names=check_duplicate_names,
                            check_capitalization=check_capitalization,
                            report_errors=True)
                            
        # flag species names where the generic name is a placeholder,
        # but the specific name is Latinized
        if check_species:
            failed_genus = []
            failed_sp_form = []
            for gid, taxa in taxonomy.items():
                species = taxa[6]
                tokens = species[3:].split()
                if len(tokens) == 2:
                    generic, specific = species[3:].split()
                    
                    canonical_generic = canonical_taxon_name('g__' + generic)
                    if placeholder_genus(canonical_generic):
                        if latinized_specific_sp_name(specific):
                            failed_genus.append((gid, species))
                else:
                    failed_sp_form.append((gid, taxa[6]))

            if failed_genus:
                print('\nTaxonomy contains species with placeholder generic names and Latinized specific names:')
                for gid, species in failed_genus:
                    print('{}\t{}'.format(gid, species))
                    
            if failed_sp_form:
                print('\nTaxonomy contains species with invalidly formed species names:')
                for gid, species in failed_sp_form:
                    print('{}\t{}'.format(gid, species))
                    
        # check that sp<GID> specific names match the NCBI accession
        if check_species:
            failed = []
            for gid, taxa in taxonomy.items():
                if gid.startswith('D-') or gid.startswith('U_') or gid.startswith('UBA'):
                    continue
                
                gid = canonical_gid(gid)
                species = taxa[6]
                tokens = species[3:].split()
                if len(tokens) == 2:
                    generic, specific = species[3:].split()
                    
                    if specific.startswith('sp') and all(ch.isdigit() for ch in specific[2:]):
                        if gid[1:] != specific[2:]:
                            failed.append((gid, species))
                        
            if failed:
                print('\nTaxonomy contains {:,} specific species identifier that do not match genome identifier, e.g.:'.format(len(failed)))
                for idx, (gid, species) in enumerate(failed):
                    if idx >= 3:
                        break
                        
                    print('{}\t{}'.format(gid, species))
                    
